"""Define tests for VELUX ACTIVE modules."""

import json
from unittest.mock import AsyncMock, patch

import pyatmo
from pyatmo import DeviceType
from pyatmo.modules.device_types import DeviceCategory
from tests.common import MockResponse, load_fixture


async def test_async_velux_modules(async_auth):
    """Test VELUX gateway and cover parsing."""
    homesdata = json.loads(load_fixture("homesdata_velux.json"))
    homestatus = json.loads(load_fixture("homestatus_velux_home_id.json"))
    homestatus["body"]["home"]["modules"][0]["name"] = "VELUX gateway"
    homestatus["body"]["home"]["modules"][0]["reachable"] = False

    account = pyatmo.AsyncAccount(async_auth)

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(
            side_effect=[
                MockResponse(homesdata, 200),
                MockResponse(homestatus, 200),
            ],
        ),
    ):
        await account.async_update_topology()
        await account.async_update_status("velux_home_id")

    home = account.homes["velux_home_id"]
    assert home.name == "VELUX Test Home"

    gateway = home.modules["velux_gateway_id"]
    assert gateway.device_type == DeviceType.NXG
    assert gateway.device_category is None
    assert gateway.wifi_strength == 44
    assert gateway.locked is True
    assert gateway.locking is False
    assert gateway.secure is True
    assert gateway.modules == ["velux_opener_awning", "velux_opener_blind"]

    opener = home.modules["velux_opener_awning"]
    assert opener.device_type == DeviceType.NXO
    assert opener.device_category == DeviceCategory.shutter
    assert opener.name == "Bedroom Awning"
    assert opener.bridge == "velux_gateway_id"
    assert opener.current_position == 100
    assert opener.target_position == 100
    assert opener.mode == "algo_disabled"
    assert opener.reachable is True
    assert opener.silent is False
    assert opener.velux_type == "awning_blind"
    assert opener.manufacturer == "Netatmo"
    assert opener.firmware_revision == 14

    blind = home.modules["velux_opener_blind"]
    assert blind.name == "Bedroom Blind"

    room = home.rooms["velux_room_id"]
    assert room.device_types == {DeviceType.NXO}


async def test_async_shutter_nxo(async_auth):
    """Test VELUX cover control payloads."""
    homesdata = json.loads(load_fixture("homesdata_velux.json"))
    homestatus = json.loads(load_fixture("homestatus_velux_home_id.json"))

    account = pyatmo.AsyncAccount(async_auth)

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(
            side_effect=[
                MockResponse(homesdata, 200),
                MockResponse(homestatus, 200),
            ],
        ),
    ):
        await account.async_update_topology()
        await account.async_update_status("velux_home_id")

    module = account.homes["velux_home_id"].modules["velux_opener_awning"]

    def gen_json_data(position):
        return {
            "json": {
                "home": {
                    "id": "velux_home_id",
                    "modules": [
                        {
                            "bridge": "velux_gateway_id",
                            "id": "velux_opener_awning",
                            "target_position": position,
                        },
                    ],
                },
            },
        }

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=MockResponse({"status": "ok"}, 200)),
    ) as mock_resp:
        assert await module.async_open()
        mock_resp.assert_awaited_with(
            params=gen_json_data(100),
            endpoint="api/setstate",
        )

        assert await module.async_close()
        mock_resp.assert_awaited_with(
            params=gen_json_data(0),
            endpoint="api/setstate",
        )

        assert await module.async_stop()
        mock_resp.assert_awaited_with(
            params=gen_json_data(-1),
            endpoint="api/setstate",
        )



"""Expose submodules."""

from .base_class import Place
from .bticino import (
    BNAB,
    BNAS,
    BNCS,
    BNCX,
    BNDL,
    BNEU,
    BNFC,
    BNIL,
    BNLD,
    BNMH,
    BNMS,
    BNSL,
    BNTH,
    BNTR,
    BNXM,
)
from .idiamant import NBG, NBO, NBR, NBS
from .legrand import (
    EBU,
    NLAO,
    NLAS,
    NLC,
    NLD,
    NLDD,
    NLE,
    NLF,
    NLFE,
    NLFN,
    NLG,
    NLIS,
    NLJ,
    NLL,
    NLLF,
    NLLM,
    NLLV,
    NLM,
    NLP,
    NLPBS,
    NLPC,
    NLPD,
    NLPM,
    NLPO,
    NLPS,
    NLPT,
    NLT,
    NLTS,
    NLUF,
    NLUI,
    NLUO,
    NLUP,
    NLV,
    NLY,
    Z3L,
    Z3V,
    NLunknown,
)
from .module import Camera, Dimmer, Fan, Module, Shutter, Switch
from .netatmo import (
    NAC,
    NCO,
    NDB,
    NHC,
    NIS,
    NOC,
    NPC,
    NRV,
    NSD,
    OTH,
    OTM,
    Location,
    NACamDoorTag,
    NACamera,
    NAMain,
    NAModule1,
    NAModule2,
    NAModule3,
    NAModule4,
    NAPlug,
    NATherm1,
    PublicWeatherArea,
)
from .smarther import BNS
from .somfy import TPSRS
from .velux import NXG, NXO

__all__ = [
    "BNAB",
    "BNAS",
    "BNCS",
    "BNCX",
    "BNDL",
    "BNEU",
    "BNFC",
    "BNIL",
    "BNLD",
    "BNMH",
    "BNMS",
    "BNS",
    "BNSL",
    "BNTH",
    "BNTR",
    "BNXM",
    "EBU",
    "NAC",
    "NBG",
    "NBO",
    "NBR",
    "NBS",
    "NCO",
    "NDB",
    "NHC",
    "NIS",
    "NLAO",
    "NLAS",
    "NLC",
    "NLD",
    "NLDD",
    "NLE",
    "NLF",
    "NLFE",
    "NLFN",
    "NLG",
    "NLIS",
    "NLJ",
    "NLL",
    "NLLF",
    "NLLM",
    "NLLV",
    "NLM",
    "NLP",
    "NLPBS",
    "NLPC",
    "NLPD",
    "NLPM",
    "NLPO",
    "NLPS",
    "NLPT",
    "NLT",
    "NLTS",
    "NLUF",
    "NLUI",
    "NLUO",
    "NLUP",
    "NLV",
    "NLY",
    "NOC",
    "NPC",
    "NRV",
    "NSD",
    "NXG",
    "NXO",
    "OTH",
    "OTM",
    "TPSRS",
    "Z3L",
    "Z3V",
    "Camera",
    "Dimmer",
    "Fan",
    "Location",
    "Module",
    "NACamDoorTag",
    "NACamera",
    "NAMain",
    "NAModule1",
    "NAModule2",
    "NAModule3",
    "NAModule4",
    "NAPlug",
    "NATherm1",
    "NLunknown",
    "Place",
    "PublicWeatherArea",
    "Shutter",
    "Switch",
]



"""Support for Netatmo authentication."""

from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import UTC, datetime
from email.utils import parsedate_to_datetime
from json import JSONDecodeError
import logging
from typing import Any

from aiohttp import (
    ClientError,
    ClientResponse,
    ClientSession,
    ClientTimeout,
    ContentTypeError,
)
from tenacity import (
    RetryCallState,
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_combine,
    wait_exponential,
    wait_random,
)

from pyatmo.const import (
    AUTHORIZATION_HEADER,
    CONCURRENCY_ERROR_CODE,
    DEFAULT_BASE_URL,
    ERRORS,
    FORBIDDEN_ERROR_CODE,
    THROTTLING_ERROR_CODE,
    TOO_MANY_REQUESTS_ERROR_CODE,
    WEBHOOK_URL_ADD_ENDPOINT,
    WEBHOOK_URL_DROP_ENDPOINT,
)
from pyatmo.exceptions import ApiError, ApiThrottlingError, ApiTooManyRequestError

LOG: logging.Logger = logging.getLogger(__name__)

# Retries to official API on 429 concurrency errors
MAX_RETRIES = 4  # total attempts
INITIAL_BACKOFF = 1  # seconds
MULTIPLIER = 1
MAX_BACKOFF = 8  # cap on a single fallback wait
MAX_RETRY_AFTER = 60  # cap on an honored server Retry-After hint


def _parse_retry_after(value: str | None) -> float | None:
    """Parse a Retry-After header (RFC 7231) into seconds.

    Accepts either a delta-seconds integer or an HTTP-date. Returns None when
    the header is absent or unparseable. A date in the past clamps to 0.
    """
    if not value:
        return None

    value = value.strip()
    if value.isdigit():
        return float(value)

    try:
        retry_date = parsedate_to_datetime(value)
    except (TypeError, ValueError):
        return None

    if retry_date.tzinfo is None:
        retry_date = retry_date.replace(tzinfo=UTC)

    delta = (retry_date - datetime.now(UTC)).total_seconds()
    return max(delta, 0.0)


# Bounded exponential backoff with jitter. wait_exponential honors min/max
# on all supported tenacity versions (wait_random_exponential only respects
# min from 9.1.0), so combine it with wait_random for the jitter.
_fallback_wait = wait_combine(
    wait_exponential(multiplier=MULTIPLIER, min=INITIAL_BACKOFF, max=MAX_BACKOFF),
    wait_random(0, 1),
)


def _wait_retry_after(retry_state: RetryCallState) -> float:
    """Wait strategy honoring a server Retry-After, else bounded backoff.

    Prefers the ``retry_after`` carried by ``ApiTooManyRequestError`` (capped
    at ``MAX_RETRY_AFTER``); otherwise falls back to a jittered, bounded
    exponential backoff.
    """
    exc = retry_state.outcome.exception() if retry_state.outcome else None
    if isinstance(exc, ApiTooManyRequestError) and exc.retry_after is not None:
        return min(exc.retry_after, MAX_RETRY_AFTER)
    return _fallback_wait(retry_state)


class AbstractAsyncAuth(ABC):
    """Abstract class to make authenticated requests."""

    def __init__(
        self,
        websession: ClientSession,
        base_url: str = DEFAULT_BASE_URL,
    ) -> None:
        """Initialize the auth."""

        self.websession: ClientSession = websession
        self.base_url: str = base_url

    @abstractmethod
    async def async_get_access_token(self) -> str:
        """Return a valid access token."""

    async def async_get_image(
        self,
        endpoint: str,
        base_url: str | None = None,
        params: dict[str, Any] | None = None,
    ) -> bytes:
        """Wrap async get requests."""

        # Note: the 429/concurrency retry lives on async_post_api_request only.
        # Camera snapshots are best-effort and time-sensitive - retrying a live
        # image seconds later has no value - so this path is deliberately not
        # decorated.
        try:
            access_token: str = await self.async_get_access_token()
        except ClientError as err:
            error_type: str = type(err).__name__
            msg: str = f"Access token failure: {error_type} - {err}"
            raise ApiError(msg) from err
        headers: dict[str, str] = {AUTHORIZATION_HEADER: f"Bearer {access_token}"}

        url: str = (base_url or self.base_url) + endpoint
        async with self.websession.get(
            url,
            params=params,
            headers=headers,
            timeout=ClientTimeout(total=5),
        ) as resp:
            resp_content: bytes = await resp.read()

            if resp.headers.get("content-type") == "image/jpeg":
                return resp_content

        msg = f"{resp.status} - invalid content-type in response when accessing '{url}'"
        raise ApiError(msg)

    @retry(
        retry=retry_if_exception_type(ApiTooManyRequestError),
        stop=stop_after_attempt(MAX_RETRIES),
        wait=_wait_retry_after,
        before_sleep=before_sleep_log(LOG, logging.DEBUG),
        reraise=True,
    )
    async def async_post_api_request(
        self,
        endpoint: str,
        base_url: str | None = None,
        params: dict[str, Any] | None = None,
    ) -> ClientResponse:
        """Wrap async post requests."""

        return await self.async_post_request(
            url=(base_url or self.base_url) + endpoint,
            params=params,
        )

    async def async_post_request(
        self,
        url: str,
        params: dict[str, Any] | None = None,
    ) -> ClientResponse:
        """Wrap async post requests."""

        access_token: str = await self.get_access_token()
        headers: dict[str, str] = {AUTHORIZATION_HEADER: f"Bearer {access_token}"}

        req_args: dict[str, Any] = self.prepare_request_arguments(params)

        async with self.websession.post(
            url,
            **req_args,
            headers=headers,
            timeout=ClientTimeout(total=5),
        ) as resp:
            return await self.process_response(resp, url)

    async def get_access_token(self) -> str:
        """Get access token."""
        try:
            return await self.async_get_access_token()
        except ClientError as err:
            msg: str = f"Access token failure: {err}"
            raise ApiError(msg) from err

    def prepare_request_arguments(self, params: dict | None) -> dict:
        """Prepare request arguments."""
        req_args: dict[str, Any] = {"data": params if params is not None else {}}

        if "params" in req_args["data"]:
            req_args["params"] = req_args["data"]["params"]
            req_args["data"].pop("params")

        if "json" in req_args["data"]:
            req_args["json"] = req_args["data"]["json"]
            req_args.pop("data")

        return req_args

    async def process_response(self, resp: ClientResponse, url: str) -> ClientResponse:
        """Process response."""
        resp_status: int = resp.status
        resp_content: bytes = await resp.read()

        if not resp.ok:
            LOG.debug("The Netatmo API returned %s (%s)", resp_content, resp_status)
            await self.handle_error_response(resp, resp_status, url)

        return await self.handle_success_response(resp, resp_content)

    async def handle_error_response(
        self,
        resp: ClientResponse,
        resp_status: int,
        url: str,
    ) -> None:
        """Handle error response."""
        try:
            resp_json: dict[str, Any] = await resp.json()
            error: dict[str, Any] = resp_json.get("error", {})
            error_code = error.get("code")

            message: str = (
                f"{resp_status} - "
                f"{ERRORS.get(resp_status, '')} - "
                f"{error.get('message')} "
                f"({error_code}) "
                f"when accessing '{url}'"
            )

            if (
                resp_status == TOO_MANY_REQUESTS_ERROR_CODE
                and error_code == CONCURRENCY_ERROR_CODE
            ):
                retry_after = _parse_retry_after(resp.headers.get("Retry-After"))
                raise ApiTooManyRequestError(message, retry_after=retry_after)

            if (
                resp_status == FORBIDDEN_ERROR_CODE
                and error_code == THROTTLING_ERROR_CODE
            ):
                raise ApiThrottlingError(message)

            raise ApiError(message)

        except (JSONDecodeError, ContentTypeError) as exc:
            msg: str = (
                f"{resp_status} - "
                f"{ERRORS.get(resp_status, '')} - "
                f"when accessing '{url}'"
            )
            raise ApiError(msg) from exc

    async def handle_success_response(
        self,
        resp: ClientResponse,
        resp_content: bytes,
    ) -> ClientResponse:
        """Handle success response."""
        try:
            if "application/json" in resp.headers.get("content-type", []):
                return resp

            if resp_content not in [b"", b"None"]:
                return resp

        except (TypeError, AttributeError):
            LOG.debug("Invalid response %s", resp)

        return resp

    async def async_addwebhook(self, webhook_url: str) -> None:
        """Register webhook."""
        try:
            resp: ClientResponse = await self.async_post_api_request(
                endpoint=WEBHOOK_URL_ADD_ENDPOINT,
                params={"url": webhook_url},
            )
        except TimeoutError as exc:
            msg: str = "Webhook registration timed out"
            raise ApiError(msg) from exc
        else:
            LOG.debug("addwebhook: %s", resp)

    async def async_dropwebhook(self) -> None:
        """Unregister webhook."""
        try:
            resp: ClientResponse = await self.async_post_api_request(
                endpoint=WEBHOOK_URL_DROP_ENDPOINT,
                params={"app_types": "app_security"},
            )
        except TimeoutError as exc:
            msg: str = "Webhook registration timed out"
            raise ApiError(msg) from exc
        else:
            LOG.debug("dropwebhook: %s", resp)



"""Define tests for fan module."""

from pyatmo import DeviceType


async def test_async_fan_NLLF(async_home):
    """Test NLLF Legrand centralized ventilation controller."""
    module_id = "12:34:56:00:01:01:01:b1"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NLLF
    assert module.power == 11
    assert module.firmware_revision == 60
    assert module.fan_speed == 1



"""WebRTC-related definitions."""

from dataclasses import dataclass


@dataclass
class WebRTCStream:
    """A WebRTC stream."""

    session_id: str
    tag_id: str


@dataclass
class WebRTCAnswer:
    """A WebRTC answer from the Netatmo API."""

    stream: WebRTCStream
    sdp: str



"""Define shared fixtures."""

from contextlib import contextmanager
from unittest.mock import AsyncMock, patch

import pytest

import pyatmo

from .common import fake_post_request, fake_post_request_multi


@contextmanager
def does_not_raise():
    yield


@pytest.fixture
async def async_auth():
    """AsyncAuth fixture."""
    with patch("pyatmo.auth.AbstractAsyncAuth", AsyncMock()) as auth:
        yield auth


@pytest.fixture
async def async_account(async_auth):
    """AsyncAccount fixture."""
    account: pyatmo.AsyncAccount = pyatmo.AsyncAccount(async_auth)

    with (
        patch(
            "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
            fake_post_request,
        ),
        patch(
            "pyatmo.auth.AbstractAsyncAuth.async_post_request",
            fake_post_request,
        ),
    ):
        await account.async_update_topology()
        yield account


@pytest.fixture
async def async_home(async_account):
    """AsyncClimate fixture for home_id 91763b24c43d3e344f424e8b."""
    home_id = "91763b24c43d3e344f424e8b"
    await async_account.async_update_status(home_id)
    return async_account.homes[home_id]


@pytest.fixture
async def async_account_multi(async_auth):
    """AsyncAccount fixture."""
    account: pyatmo.AsyncAccount = pyatmo.AsyncAccount(async_auth)

    with (
        patch(
            "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
            fake_post_request_multi,
        ),
        patch(
            "pyatmo.auth.AbstractAsyncAuth.async_post_request",
            fake_post_request_multi,
        ),
    ):
        await account.async_update_topology(
            disabled_homes_ids=["eeeeeeeeeffffffffffaaaaa"],
        )
        yield account


@pytest.fixture
async def async_home_multi(async_account_multi):
    """AsyncClimate fixture for home_id 91763b24c43d3e344f424e8b."""
    home_id = "aaaaaaaaaaabbbbbbbbbbccc"
    return async_account_multi.homes[home_id]



"""Define tests for the module base helpers."""

import pytest

from pyatmo.modules.module import process_battery_state


@pytest.mark.parametrize(
    ("battery_state", "expected"),
    [
        ("max", 100),
        ("full", 90),
        ("high", 75),
        ("medium", 50),
        ("low", 25),
        ("very_low", 10),
    ],
)
def test_process_battery_state_known_values(battery_state: str, expected: int) -> None:
    """Known battery states map to their percent value."""

    assert process_battery_state(battery_state) == expected


def test_process_battery_state_unknown_value_returns_zero() -> None:
    """An unknown battery state degrades gracefully to 0 instead of raising."""

    assert process_battery_state("empty") == 0



"""Define tests for home module."""

import json
import logging
from unittest.mock import AsyncMock, patch

import anyio
import pytest

import pyatmo
from pyatmo import DeviceType, InvalidScheduleError, NoDeviceError
from pyatmo.enums import (
    SCHEDULE_TYPE_MAPPING,
    PressureUnit,
    ScheduleType,
    TemperatureControlMode,
    UnitSystem,
    WindUnit,
)
from pyatmo.home import Home, get_temperature_control_mode
from tests.common import MockResponse


async def test_async_home(async_home):
    """Test basic home setup."""
    room_id = "3688132631"
    room = async_home.rooms[room_id]
    assert room.device_types == {
        DeviceType.NDB,
        DeviceType.NACamera,
        DeviceType.NBR,
        DeviceType.NIS,
        DeviceType.NBO,
        DeviceType.NPC,
        DeviceType.NLPD,
    }
    assert len(async_home.rooms) == 9
    assert len(async_home.modules) == 51
    assert async_home.modules != room.modules

    module_id = "12:34:56:10:f1:66"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NDB

    module_id = "12:34:56:10:b9:0e"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NOC

    assert async_home.temperature_control_mode == "cooling"


async def test_async_home_set_schedule(async_home):
    """Test home schedule."""
    schedule_id = "591b54a2764ff4d50d8b5795"
    selected_schedule = async_home.get_selected_schedule()
    assert selected_schedule.entity_id == schedule_id
    assert async_home.is_valid_schedule(schedule_id)
    assert not async_home.is_valid_schedule("123")
    assert async_home.get_hg_temp() == 7
    assert async_home.get_away_temp() == 14


async def test_async_set_schedule_temperatures(async_home):
    """Test setting schedule temperatures."""
    schedule_id = "591b54a2764ff4d50d8b5795"
    schedule = async_home.get_selected_schedule()

    assert schedule.entity_id == schedule_id
    zone = next((zone for zone in schedule.zones if zone.entity_id == 1), None)
    assert zone is not None
    room = next((room for room in zone.rooms if room.entity_id == "2746182631"), None)
    assert room is not None
    assert room.therm_setpoint_temperature == 17

    temps = {"2746182631": 15}

    async with await anyio.open_file(
        "fixtures/sync_schedule_591b54a2764ff4d50d8b5795.json",
        encoding="utf-8",
    ) as fixture_file:
        json_fixture = json.loads(await fixture_file.read())

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=MockResponse({"status": "ok"}, 200)),
    ) as mock_resp:
        await async_home.async_set_schedule_temperatures(1, temps)

        mock_resp.assert_awaited_with(
            endpoint="api/synchomeschedule",
            params={
                "params": {
                    "home_id": "91763b24c43d3e344f424e8b",
                    "schedule_id": schedule_id,
                    "name": "Default",
                },
                "json": json_fixture,
            },
        )


async def test_async_sync_schedule(async_home):
    """Test setting schedule temperatures."""
    schedule_id = "b1b54a2f45795764f59d50d8"
    schedule = async_home.schedules.get(schedule_id)

    assert schedule is not None
    assert schedule.entity_id == schedule_id
    zone = next((zone for zone in schedule.zones if zone.entity_id == 1), None)
    assert zone is not None
    room = next((room for room in zone.rooms if room.entity_id == "2746182631"), None)
    assert room is not None
    assert room.therm_setpoint_temperature == 17

    # set a new room temperature
    room.therm_setpoint_temperature = 14

    async with await anyio.open_file(
        "fixtures/sync_schedule_b1b54a2f45795764f59d50d8.json",
        encoding="utf-8",
    ) as fixture_file:
        json_fixture = json.loads(await fixture_file.read())

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=MockResponse({"status": "ok"}, 200)),
    ) as mock_resp:
        await async_home.async_sync_schedule(schedule)

        mock_resp.assert_awaited_with(
            endpoint="api/synchomeschedule",
            params={
                "params": {
                    "home_id": "91763b24c43d3e344f424e8b",
                    "schedule_id": schedule_id,
                    "name": "Default",
                },
                "json": json_fixture,
            },
        )


async def test_async_sync_schedule_invalid_schedule(async_home):
    """Test syncing an invalid schedule."""
    invalid_schedule = {"invalid": "data"}

    with (
        pytest.raises(InvalidScheduleError),
        patch(
            "pyatmo.home.is_valid_schedule",
            return_value=False,
        ),
    ):
        await async_home.async_sync_schedule(invalid_schedule)


async def test_async_home_data_no_body(async_auth):
    """Test home data with no body."""
    async with await anyio.open_file(
        "fixtures/homesdata_empty.json",
        encoding="utf-8",
    ) as fixture_file:
        json_fixture = json.loads(await fixture_file.read())

    mock_request = AsyncMock(return_value=MockResponse(json_fixture, 200))
    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        mock_request,
    ):
        climate = pyatmo.AsyncAccount(async_auth)
        with pytest.raises(NoDeviceError):
            await climate.async_update_topology()
        mock_request.assert_awaited_once()


async def test_async_set_persons_home(async_account):
    """Test marking a person being at home."""
    home_id = "91763b24c43d3e344f424e8b"
    home = async_account.homes[home_id]

    person_ids = [
        "91827374-7e04-5298-83ad-a0cb8372dff1",
        "91827375-7e04-5298-83ae-a0cb8372dff2",
    ]

    async with await anyio.open_file(
        "fixtures/status_ok.json",
        encoding="utf-8",
    ) as json_file:
        response = json.loads(await json_file.read())

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=MockResponse(response, 200)),
    ) as mock_resp:
        await home.async_set_persons_home(person_ids)

        mock_resp.assert_awaited_with(
            params={"home_id": home_id, "person_ids[]": person_ids},
            endpoint="api/setpersonshome",
        )


async def test_async_set_persons_away(async_account):
    """Test marking a set of persons being away."""
    home_id = "91763b24c43d3e344f424e8b"
    home = async_account.homes[home_id]

    async with await anyio.open_file(
        "fixtures/status_ok.json",
        encoding="utf-8",
    ) as json_file:
        response = json.loads(await json_file.read())

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=MockResponse(response, 200)),
    ) as mock_resp:
        person_id = "91827374-7e04-5298-83ad-a0cb8372dff1"
        await home.async_set_persons_away(person_id)

        mock_resp.assert_awaited_with(
            params={"home_id": home_id, "person_id": person_id},
            endpoint="api/setpersonsaway",
        )

        await home.async_set_persons_away()

        mock_resp.assert_awaited_with(
            params={"home_id": home_id},
            endpoint="api/setpersonsaway",
        )


async def test_home_event_update(async_account):
    """Test basic event update."""
    home_id = "91763b24c43d3e344f424e8b"
    await async_account.async_update_events(home_id=home_id)
    home = async_account.homes[home_id]

    events = home.events
    assert len(events) == 8

    module_id = "12:34:56:10:b9:0e"
    assert module_id in home.modules
    module = home.modules[module_id]

    events = module.events
    assert len(events) == 5
    assert events[0].event_type == "outdoor"
    assert events[0].video_id == "11111111-2222-3333-4444-b42f0fc4cfad"
    assert events[1].event_type == "connection"


async def test_async_home_module_error_code(async_account):
    """Test that per-module error code from homestatus errors[] is surfaced."""
    home_id = "91763b24c43d3e344f424e8b"
    await async_account.async_update_status(home_id)
    home = async_account.homes[home_id]

    module_id = "12:34:56:00:fa:d0"
    assert module_id in home.modules
    module = home.modules[module_id]

    async with await anyio.open_file(
        "fixtures/home_status_error_disconnected.json",
        encoding="utf-8",
    ) as json_file:
        home_status_fixture = json.loads(await json_file.read())
    mock_home_status_resp = MockResponse(home_status_fixture, 200)

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=mock_home_status_resp),
    ) as mock_request:
        await async_account.async_update_status(home_id)
        mock_request.assert_called()

    assert module.error_code == 6
    assert "error_code" not in module.features

    # Recovery: a subsequent healthy /homestatus update (the module is reported
    # in home.modules again) must clear the stale error code back to None.
    await async_account.async_update_status(home_id)
    assert module.error_code is None


async def test_async_home_module_setup_date(async_home):
    """Test that module setup_date from /homesdata topology is surfaced."""
    module_id = "12:34:56:00:fa:d0"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.setup_date == 1494963356
    assert "setup_date" not in module.features


async def test_async_home_room_type_and_therm_relay(async_home):
    """Test room type + therm_relay from /homesdata topology are surfaced."""
    livingroom = async_home.rooms["2746182631"]
    assert livingroom.room_type == "livingroom"

    bureau = async_home.rooms["222452125"]
    assert bureau.room_type == "electrical_cabinet"
    assert bureau.therm_relay == "12:34:56:20:f5:44"

    # Rooms without therm_relay in the payload default to None.
    assert livingroom.therm_relay is None


async def test_async_account_user_country_and_consent(async_account):
    """Test user country + pending_user_consent from /homesdata are surfaced."""
    assert async_account.user == "john@doe.com"
    assert async_account.user_country == "DE"
    assert async_account.pending_user_consent is True

    # The raw user block (email/id PII) must not leak into raw_data, which
    # consumers such as Home Assistant diagnostics serialize wholesale.
    assert "user" not in async_account.raw_data


async def test_async_home_electricity_schedule(async_home):
    """Test electricity schedule tariff fields + zone prices are parsed."""
    schedule = async_home.schedules["c1c54a2f45795764f59d50d9"]
    assert schedule.type == "electricity"
    assert schedule.tariff == "custom"
    assert schedule.tariff_option == "peak_and_off_peak"
    assert schedule.power_threshold == 6
    assert schedule.contract_power_unit == "kVA"

    peak = next(z for z in schedule.zones if z.entity_id == 0)
    assert peak.price_type == "peak"
    assert peak.price_value == 0.21
    off_peak = next(z for z in schedule.zones if z.entity_id == 1)
    assert off_peak.price_type == "off_peak"
    assert off_peak.price_value == 0.16


async def test_async_home_event_schedule(async_home):
    """Test event schedule twilight timetables + zone module actions are parsed."""
    schedule = async_home.schedules["d1d54a2f45795764f59d50da"]
    assert schedule.type == "event"

    assert len(schedule.timetable_sunrise) == 1
    sunrise = schedule.timetable_sunrise[0]
    assert sunrise.zone_id == 0
    assert sunrise.day == 1
    assert sunrise.twilight_offset == -30

    assert len(schedule.timetable_sunset) == 1
    assert schedule.timetable_sunset[0].twilight_offset == 15

    morning = next(z for z in schedule.zones if z.entity_id == 0)
    assert len(morning.modules) == 1
    light = morning.modules[0]
    assert light.entity_id == "12:34:56:00:01:ae"
    assert light.bridge == "12:34:56:00:fa:d0"
    assert light.on is True
    assert light.brightness == 80

    evening = next(z for z in schedule.zones if z.entity_id == 1)
    shutter = evening.modules[0]
    assert shutter.target_position == 100
    assert shutter.fan_speed == 2


def test_device_types_missing():
    """Test handling of missing device types."""

    assert DeviceType("NOC") == DeviceType.NOC
    assert DeviceType("UNKNOWN") == DeviceType.NLunknown


async def test_module_bridged_key_variants(async_home):
    """Both `modules_bridged` and `module_bridged` populate Module.modules."""
    plural = async_home.get_module(
        {"id": "aa:aa", "type": "NLP", "modules_bridged": ["child-1"]},
    )
    assert plural.modules == ["child-1"]

    singular = async_home.get_module(
        {"id": "bb:bb", "type": "NLP", "module_bridged": ["child-2"]},
    )
    assert singular.modules == ["child-2"]


async def test_module_bridged_key_precedence(async_home):
    """When both keys are present, `modules_bridged` wins over `module_bridged`."""
    mod = async_home.get_module(
        {
            "id": "ee:ee",
            "type": "NLP",
            "modules_bridged": ["canonical"],
            "module_bridged": ["alias"],
        },
    )
    assert mod.modules == ["canonical"]


async def test_module_bridged_key_topology_update(async_home):
    """Both bridged-key spellings work on the update_topology reflection path."""
    # update_topology -> _update_attributes drives the NETATMO_ATTRIBUTES_MAP
    # "modules" entry (bridged_module_ids), a different code path than
    # Module.__init__.
    plural = async_home.get_module({"id": "cc:cc", "type": "NLP"})
    plural.update_topology(
        {"id": "cc:cc", "type": "NLP", "modules_bridged": ["child-3"]},
    )
    assert plural.modules == ["child-3"]

    singular = async_home.get_module({"id": "dd:dd", "type": "NLP"})
    singular.update_topology(
        {"id": "dd:dd", "type": "NLP", "module_bridged": ["child-4"]},
    )
    assert singular.modules == ["child-4"]


async def test_home_geolocation(async_home):
    """Test home geolocation fields from /homesdata topology are surfaced."""
    assert async_home.altitude == 112
    assert async_home.country == "DE"
    assert async_home.timezone == "Europe/Berlin"
    # Coordinates are exposed as-is; the API's lat/lon ordering is ambiguous
    # across schema variants, so the library does not reinterpret them.
    assert async_home.coordinates == [52.516263, 13.377726]


async def test_home_geolocation_topology_update(async_home):
    """Geolocation is refreshed on the update_topology path, not only __init__."""
    async_home.update_topology(
        {
            "id": async_home.entity_id,
            "name": "MYHOME",
            "altitude": 5,
            "coordinates": [1.0, 2.0],
            "country": "FR",
            "timezone": "Europe/Paris",
        },
    )
    assert async_home.altitude == 5
    assert async_home.coordinates == [1.0, 2.0]
    assert async_home.country == "FR"
    assert async_home.timezone == "Europe/Paris"


async def test_home_geolocation_partial_update_preserves(async_home):
    """A partial topology update must not wipe known geolocation.

    Home.update calls update_topology({"modules": [...]}) for newly-seen
    modules; that payload omits the geolocation keys and must keep the
    values already populated from /homesdata.
    """
    assert async_home.altitude == 112
    assert async_home.coordinates == [52.516263, 13.377726]

    async_home.update_topology({"modules": []})

    assert async_home.altitude == 112
    assert async_home.coordinates == [52.516263, 13.377726]
    assert async_home.country == "DE"
    assert async_home.timezone == "Europe/Berlin"


async def test_home_update_new_module_preserves_home_fields(async_home):
    """Discovering a new module via /homestatus must not wipe home fields.

    Home.update registers a not-yet-seen module; that path must not reset
    name/therm state/geolocation, whose keys are absent from /homestatus.
    """
    name = async_home.name
    altitude = async_home.altitude
    coordinates = async_home.coordinates
    therm_mode = async_home.therm_mode
    assert name == "MYHOME"

    # Capture the pre-existing modules: routing a single-module payload through
    # update_topology used to pop every other module via its removal loop.
    existing_module_ids = set(async_home.modules)
    assert len(existing_module_ids) > 1

    new_module = {"id": "ff:ff:ff:ff:ff:ff", "type": "NAMain"}
    assert new_module["id"] not in async_home.modules

    await async_home.update(
        {"home": {"id": async_home.entity_id, "modules": [new_module]}},
    )

    assert new_module["id"] in async_home.modules
    assert existing_module_ids <= set(async_home.modules)
    assert async_home.name == name
    assert async_home.altitude == altitude
    assert async_home.coordinates == coordinates
    assert async_home.therm_mode == therm_mode


async def test_account_user_units(async_account):
    """Test user display-unit preferences from /homesdata are surfaced."""
    assert async_account.unit_system == UnitSystem.METRIC
    assert async_account.unit_wind == WindUnit.MPH
    assert async_account.unit_pressure == PressureUnit.MMHG

    # The raw user block (email/id PII) must not leak into raw_data.
    assert "user" not in async_account.raw_data


async def test_account_user_units_missing(async_auth):
    """Test unit preferences stay None when the keys are absent."""
    account = pyatmo.AsyncAccount(async_auth)

    async def fake_request(*_, **__):
        body = {"body": {"homes": [{"id": "1", "name": "home"}], "user": {}}}
        return MockResponse(body, 200)

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        fake_request,
    ):
        await account.async_update_topology()

    assert account.unit_system is None
    assert account.unit_wind is None
    assert account.unit_pressure is None


def test_account_user_units_unknown_fallback(caplog):
    """Test out-of-range unit codes fall back to UNKNOWN and log a warning."""
    assert UnitSystem(99) is UnitSystem.UNKNOWN
    assert WindUnit(99) is WindUnit.UNKNOWN
    assert PressureUnit(99) is PressureUnit.UNKNOWN
    assert "unknown" in caplog.text.lower()


async def test_device_type_aliases(async_home):
    """Legacy/typo module types from /homesdata resolve to canonical classes."""
    # `NBD` is a transposition of the canonical `NDB` (Smart Video Doorbell).
    doorbell = async_home.get_module({"id": "1", "type": "NBD"})
    assert isinstance(doorbell, pyatmo.modules.NDB)
    assert doorbell.device_type == DeviceType.NDB

    # `NADoorTag` is a legacy alias of `NACamDoorTag`.
    doortag = async_home.get_module({"id": "2", "type": "NADoorTag"})
    assert isinstance(doortag, pyatmo.modules.NACamDoorTag)
    assert doortag.device_type == DeviceType.NACamDoorTag


@pytest.mark.parametrize(
    ("raw_value", "expected"),
    [
        ("heating", TemperatureControlMode.HEATING),
        ("cooling", TemperatureControlMode.COOLING),
        ("auto", TemperatureControlMode.AUTO),
        (None, None),
    ],
)
def test_get_temperature_control_mode(raw_value, expected):
    """Known modes parse to their enum member; missing value stays None."""
    assert get_temperature_control_mode(raw_value) == expected


def test_get_temperature_control_mode_unknown(caplog):
    """Unknown mode degrades to None with a warning, does not raise."""
    with caplog.at_level(logging.WARNING):
        assert get_temperature_control_mode("garbage") is None
    assert "garbage" in caplog.text


def test_schedule_type_mapping_covers_all_modes():
    """Every TemperatureControlMode maps to a ScheduleType (no KeyError path)."""
    for mode in TemperatureControlMode:
        assert mode in SCHEDULE_TYPE_MAPPING
    assert SCHEDULE_TYPE_MAPPING[TemperatureControlMode.AUTO] == ScheduleType.AUTO


async def test_schedule_lookup_auto_mode_no_keyerror(async_home):
    """Home in auto mode: schedule lookups must not raise KeyError."""
    async_home.temperature_control_mode = TemperatureControlMode.AUTO
    # No auto schedule in the fixture -> no match, but must not crash.
    assert async_home.get_selected_schedule() is None
    assert async_home.get_available_schedules() == []


async def test_home_init_auto_mode_parses_and_selects_schedule(async_auth):
    """Home built from a tcm='auto' payload must not raise (issue #176631).

    Reproduces the original crash path (Home.__init__ ->
    get_temperature_control_mode) with a real raw payload rather than a
    mutated fixture, and confirms an auto-typed schedule is selectable.
    """
    raw_data = {
        "id": "auto-home",
        "name": "Auto Home",
        "temperature_control_mode": "auto",
        "schedules": [
            {
                "id": "sched-auto",
                "name": "Auto schedule",
                "type": "auto",
                "selected": True,
                "hg_temp": 7,
                "away_temp": 14,
            },
            {
                "id": "sched-therm",
                "name": "Therm schedule",
                "type": "therm",
                "selected": False,
            },
        ],
    }

    home = Home(async_auth, raw_data)

    assert home.temperature_control_mode == TemperatureControlMode.AUTO
    selected = home.get_selected_schedule()
    assert selected is not None
    assert selected.entity_id == "sched-auto"
    assert home.get_hg_temp() == 7
    assert home.get_away_temp() == 14
    assert [s.entity_id for s in home.get_available_schedules()] == ["sched-auto"]



"""Module to represent a Netatmo schedule."""

from __future__ import annotations

from dataclasses import dataclass
import logging
from typing import TYPE_CHECKING

from pyatmo.enums import ScheduleType
from pyatmo.modules.base_class import NetatmoBase
from pyatmo.room import Room

if TYPE_CHECKING:
    from pyatmo.const import RawData

    from .home import Home

LOG: logging.Logger = logging.getLogger(__name__)


@dataclass
class Schedule(NetatmoBase):
    """Class to represent a Netatmo schedule."""

    type: ScheduleType
    away_temp: float | None
    hg_temp: float | None
    cooling_away_temp: float | None
    timetable: list[TimetableEntry]
    selected: bool
    default: bool

    # electricity / electricity_production schedule fields
    tariff: str | None
    tariff_option: str | None
    power_threshold: int | None
    contract_power_unit: str | None

    # event schedule twilight timetables
    timetable_sunrise: list[TwilightEntry]
    timetable_sunset: list[TwilightEntry]

    def __init__(self, home: Home, raw_data: RawData) -> None:
        """Initialize a Netatmo schedule instance."""
        super().__init__(raw_data)
        self.home = home
        self.type = ScheduleType(raw_data.get("type", ScheduleType.THERM))
        self.default = raw_data.get("default", False)
        self.selected = raw_data.get("selected", False)
        self.hg_temp = raw_data.get("hg_temp")
        self.away_temp = raw_data.get("away_temp")
        self.cooling_away_temp = raw_data.get("cooling_away_temp")
        self.tariff = raw_data.get("tariff")
        self.tariff_option = raw_data.get("tariff_option")
        self.power_threshold = raw_data.get("power_threshold")
        self.contract_power_unit = raw_data.get("contract_power_unit")
        self.timetable = [
            TimetableEntry(home, r) for r in raw_data.get("timetable", [])
        ]
        self.timetable_sunrise = [
            TwilightEntry(home, r) for r in raw_data.get("timetable_sunrise", [])
        ]
        self.timetable_sunset = [
            TwilightEntry(home, r) for r in raw_data.get("timetable_sunset", [])
        ]
        self.zones = [Zone(home, r) for r in raw_data.get("zones", [])]

    def update_topology(self, raw_data: RawData) -> None:
        """Update the schedule topology."""
        super().update_topology(raw_data)

        self.selected = raw_data.get("selected", self.selected)
        self.default = raw_data.get("default", self.default)
        self.hg_temp = raw_data.get("hg_temp", self.hg_temp)
        self.away_temp = raw_data.get("away_temp", self.away_temp)
        self.cooling_away_temp = raw_data.get(
            "cooling_away_temp",
            self.cooling_away_temp,
        )
        self.tariff = raw_data.get("tariff", self.tariff)
        self.tariff_option = raw_data.get("tariff_option", self.tariff_option)
        self.power_threshold = raw_data.get("power_threshold", self.power_threshold)
        self.contract_power_unit = raw_data.get(
            "contract_power_unit",
            self.contract_power_unit,
        )
        self.timetable = [
            TimetableEntry(self.home, r) for r in raw_data.get("timetable", [])
        ]
        self.timetable_sunrise = [
            TwilightEntry(self.home, r) for r in raw_data.get("timetable_sunrise", [])
        ]
        self.timetable_sunset = [
            TwilightEntry(self.home, r) for r in raw_data.get("timetable_sunset", [])
        ]

        self.zones = [Zone(self.home, r) for r in raw_data.get("zones", [])]


@dataclass
class TimetableEntry:
    """Class to represent a Netatmo schedule's timetable entry."""

    zone_id: int | None
    m_offset: int | None

    def __init__(self, home: Home, raw_data: RawData) -> None:
        """Initialize a Netatmo schedule's timetable entry instance."""
        self.home = home
        self.zone_id = raw_data.get("zone_id", 0)
        self.m_offset = raw_data.get("m_offset", 0)


@dataclass
class TwilightEntry:
    """Class to represent an event schedule's sunrise/sunset timetable entry."""

    zone_id: int | None
    day: int | None
    twilight_offset: int | None

    def __init__(self, home: Home, raw_data: RawData) -> None:
        """Initialize a Netatmo schedule's twilight timetable entry instance."""
        self.home = home
        self.zone_id = raw_data.get("zone_id", 0)
        self.day = raw_data.get("day")
        self.twilight_offset = raw_data.get("twilight_offset")


@dataclass
class ZoneModule(NetatmoBase):
    """Class to represent an event schedule zone's per-module action."""

    bridge: str | None
    on: bool | None
    target_position: int | None
    brightness: int | None
    fan_speed: int | None

    def __init__(self, home: Home, raw_data: RawData) -> None:
        """Initialize a Netatmo schedule zone's module action instance."""
        super().__init__(raw_data)
        self.home = home
        self.bridge = raw_data.get("bridge")
        self.on = raw_data.get("on")
        self.target_position = raw_data.get("target_position")
        self.brightness = raw_data.get("brightness")
        self.fan_speed = raw_data.get("fan_speed")


@dataclass
class Zone(NetatmoBase):
    """Class to represent a Netatmo schedule's zone."""

    type: int
    rooms: list[Room]

    # electricity schedule zone fields
    price_type: str | None
    price_value: float | None

    # event schedule zone module actions
    modules: list[ZoneModule]

    def __init__(self, home: Home, raw_data: RawData) -> None:
        """Initialize a Netatmo schedule's zone instance."""
        super().__init__(raw_data)
        self.home = home
        self.type = raw_data.get("type", 0)
        self.price_type = raw_data.get("price_type")
        self.price_value = raw_data.get("price_value")
        self.modules = [ZoneModule(home, r) for r in raw_data.get("modules", [])]

        def room_factory(home: Home, room_raw_data: RawData) -> Room:
            room: Room = Room(home, room_raw_data, {})
            room.update(room_raw_data)
            return room

        self.rooms = [room_factory(home, r) for r in raw_data.get("rooms", [])]



"""Define tests for switch module."""

from pyatmo import DeviceType
from pyatmo.modules.device_types import DeviceCategory


async def test_async_switch_NLP(async_home):
    """Test NLP Legrand plug."""
    module_id = "12:34:56:80:00:12:ac:f2"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NLP
    assert module.firmware_revision == 62
    assert module.on
    assert module.power == 0


async def test_offload_meters(async_home):
    """offload_meters (smart-shedder ids) are parsed on offload modules."""
    module = async_home.modules["12:34:56:80:00:12:ac:f2"]
    assert module.offload_meters == ["98:76:54:32:10:ab"]


async def test_offload_meters_absent(async_home):
    """offload_meters is None when the field is absent from the response."""
    module = async_home.modules["12:34:56:00:01:01:01:b7"]
    assert module.offload_meters is None


async def test_offload_meters_empty(async_home):
    """offload_meters is an empty list when the API reports no shedders."""
    module = async_home.modules["12:34:56:80:60:40"]
    assert module.offload_meters == []


async def test_async_switch_NLF(async_home):
    """Test NLF Legrand dimmer."""
    module_id = "00:11:22:33:00:11:45:fe"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NLF
    assert module.firmware_revision == 57
    assert module.on is False
    assert module.brightness == 63
    assert module.power == 0


async def test_async_switch_NLIS(async_home):
    """Test NLIS Legrand module."""
    module_id = "12:34:56:00:01:01:01:b6"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_category is None
    module_id = "12:34:56:00:01:01:01:b6#1"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_category == DeviceCategory.switch
    module_id = "12:34:56:00:01:01:01:b6#2"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_category == DeviceCategory.switch



"""Define tests for the account module."""

from pyatmo import modules


async def test_update_devices_unknown_type_falls_back_to_nlunknown(async_account):
    """Test that an unknown standalone device type does not abort the update.

    A getstationsdata/gethomecoachsdata device reporting a ``type`` with no
    matching class in ``pyatmo.modules`` must fall back to ``NLunknown`` instead
    of raising ``AttributeError`` and aborting the whole update.
    """
    device_id = "00:11:22:33:44:55"
    device_data = {
        "_id": device_id,
        "type": "NOSUCHTYPE",
    }

    # Must not raise even though "NOSUCHTYPE" has no matching class.
    await async_account.update_devices({"devices": [device_data]})

    assert device_id in async_account.modules
    assert isinstance(async_account.modules[device_id], modules.NLunknown)



"""Module to represent a Netatmo room."""

from __future__ import annotations

from dataclasses import dataclass
import logging
from typing import TYPE_CHECKING, Any, cast

from pyatmo.const import (
    AWAY,
    COOLING,
    FROSTGUARD,
    HEATING,
    HOME,
    IDLE,
    MANUAL,
    MAX,
    OFF,
    PILOT_WIRE_AWAY,
    PILOT_WIRE_COMFORT,
    PILOT_WIRE_COMFORT_1,
    PILOT_WIRE_COMFORT_2,
    PILOT_WIRE_FROST_GUARD,
    PILOT_WIRE_STAND_BY,
    SCHEDULE,
    SETROOMTHERMPOINT_ENDPOINT,
    UNKNOWN,
    RawData,
)
from pyatmo.enums import TemperatureControlMode
from pyatmo.modules.base_class import NetatmoBase
from pyatmo.modules.device_types import ApplianceType, DeviceCategory, DeviceType
from pyatmo.modules.module import ApplianceTypeMixin, Boiler, PowerMixin

if TYPE_CHECKING:
    from pyatmo.home import Home
    from pyatmo.modules.module import Module

LOG: logging.Logger = logging.getLogger(__name__)

MODE_MAP: dict[str, str] = {SCHEDULE: HOME}

# Climate setpoint mode -> pilot-wire ("fil pilote") preset.
# Many-to-one: several modes collapse onto PILOT_WIRE_COMFORT.
_CLIMATE_SETPOINT_MODE_TO_PILOT_WIRE: dict[str, str] = {
    MANUAL: PILOT_WIRE_COMFORT,
    MAX: PILOT_WIRE_COMFORT,
    OFF: PILOT_WIRE_FROST_GUARD,
    HOME: PILOT_WIRE_COMFORT,  # HOME is equivalent to a schedule for Netatmo NLC based room
    FROSTGUARD: PILOT_WIRE_FROST_GUARD,
    SCHEDULE: PILOT_WIRE_COMFORT,
    AWAY: PILOT_WIRE_AWAY,
}

# Pilot-wire preset -> canonical NLC climate setpoint mode. Not the inverse of
# the map above: it picks one canonical mode per preset and also covers presets
# the forward map never emits (COMFORT_1, COMFORT_2, STAND_BY) that can arrive
# straight from the API.
_PILOT_WIRE_TO_CLIMATE_SETPOINT_MODE: dict[str, str] = {
    PILOT_WIRE_COMFORT: MANUAL,
    PILOT_WIRE_AWAY: MANUAL,  # AWAY is like ECO for a pilot wire heater, so put manual to force it to happen
    PILOT_WIRE_FROST_GUARD: FROSTGUARD,
    PILOT_WIRE_STAND_BY: FROSTGUARD,
    PILOT_WIRE_COMFORT_1: HOME,
    PILOT_WIRE_COMFORT_2: HOME,
}


def climate_setpoint_mode_to_pilot_wire(mode: str) -> str:
    """Map a climate setpoint mode to its pilot-wire preset (frost guard default)."""
    return _CLIMATE_SETPOINT_MODE_TO_PILOT_WIRE.get(mode, PILOT_WIRE_FROST_GUARD)


def pilot_wire_to_climate_setpoint_mode(pilot_wire: str) -> str:
    """Map a pilot-wire preset back to the canonical NLC setpoint mode (frost guard default)."""
    return _PILOT_WIRE_TO_CLIMATE_SETPOINT_MODE.get(pilot_wire, FROSTGUARD)


@dataclass
class Room(NetatmoBase):
    """Class to represent a Netatmo room."""

    modules: dict[str, Module]
    device_types: set[DeviceType]
    features: set[str]

    climate_type: DeviceType | None = None

    humidity: int | None = None
    therm_measured_temperature: float | None = None

    reachable: bool | None = None

    heating_power_request: int | None = None
    therm_setpoint_temperature: float | None = None

    therm_setpoint_mode: str | None = None
    therm_setpoint_fp: str | None = None
    support_pilot_wire: bool = False

    therm_setpoint_start_time: int | None = None
    therm_setpoint_end_time: int | None = None

    anticipating: bool | None = None
    open_window: bool | None = None

    cooling_setpoint_temperature: float | None = None
    cooling_setpoint_start_time: int | None = None
    cooling_setpoint_end_time: int | None = None
    cooling_setpoint_mode: str | None = None

    radiators_power: int | None = None

    room_type: str | None = None  # API "type": kitchen, bedroom, livingroom, ...
    therm_relay: str | None = None  # main device id of the controlling heating module

    def __init__(
        self,
        home: Home,
        room: dict[str, Any],
        all_modules: dict[str, Module],
    ) -> None:
        """Initialize a Netatmo room instance."""

        super().__init__(room)
        self.home = home
        self.support_pilot_wire = False
        self.room_type = room.get("type")
        self.therm_relay = room.get("therm_relay")
        self.modules = {
            m_id: m
            for m_id, m in all_modules.items()
            if m_id in room.get("module_ids", [])
        }
        self.device_types = set()
        self.features = set()
        self.evaluate_device_type()

    def update_topology(self, raw_data: RawData) -> None:
        """Update room topology."""

        self.name = raw_data.get("name", UNKNOWN)
        self.room_type = raw_data.get("type", self.room_type)
        self.therm_relay = raw_data.get("therm_relay", self.therm_relay)
        self.modules = {
            m_id: m
            for m_id, m in self.home.modules.items()
            if m_id in raw_data.get("module_ids", [])
        }
        self.evaluate_device_type()

    def evaluate_device_type(self) -> None:
        """Evaluate the device type of the room."""

        self.support_pilot_wire = False
        for module in self.modules.values():
            self.device_types.add(module.device_type)
            if module.device_category is not None:
                self.features.add(module.device_category.name)
            if (
                module.device_type == "NLC"
                and isinstance(module, ApplianceTypeMixin)
                and module.appliance_type == ApplianceType.radiator
            ):
                self.support_pilot_wire = True
                # Regarding to the room the cable outlet can be seen as climate control, add the climate feature
                self.features.add(DeviceCategory.climate.name)

        if "OTM" in self.device_types:
            self.climate_type = DeviceType.OTM
        elif "NATherm1" in self.device_types:
            self.climate_type = DeviceType.NATherm1
        elif "BNS" in self.device_types:
            self.climate_type = DeviceType.BNS
            self.features.add("humidity")
        elif "NRV" in self.device_types:
            self.climate_type = DeviceType.NRV
        elif "NAC" in self.device_types:
            self.climate_type = DeviceType.NAC
        elif "BNTH" in self.device_types:
            self.climate_type = DeviceType.BNTH
        elif "NLC" in self.device_types and self.support_pilot_wire:
            self.climate_type = DeviceType.NLC

    def update(self, raw_data: RawData) -> None:
        """Update room data."""

        self.humidity = raw_data.get("humidity")
        self.radiators_power = 0

        if self.climate_type == DeviceType.BNTH:
            # BNTH is wired, so the room is always reachable
            self.reachable = True
        elif self.climate_type == DeviceType.NLC:
            self.reachable = raw_data.get("reachable", False)
            for module in self.modules.values():
                if (
                    isinstance(module, ApplianceTypeMixin)
                    and module.device_type == DeviceType.NLC
                    and module.appliance_type == ApplianceType.radiator
                ):
                    if isinstance(module, PowerMixin) and module.power is not None:
                        self.radiators_power += module.power

                    if hasattr(module, "reachable"):
                        state = module.reachable
                        if state is not None:
                            if self.reachable is None:
                                self.reachable = state
                            else:
                                self.reachable = (
                                    self.reachable or state
                                )  # as soon as we do have one
        else:
            self.reachable = raw_data.get("reachable")

        self.therm_measured_temperature = raw_data.get("therm_measured_temperature")

        self.heating_power_request = raw_data.get("heating_power_request")
        self.therm_setpoint_mode = raw_data.get("therm_setpoint_mode")
        self.therm_setpoint_fp = raw_data.get("therm_setpoint_fp")
        self.therm_setpoint_temperature = raw_data.get("therm_setpoint_temperature")
        self.therm_setpoint_start_time = raw_data.get("therm_setpoint_start_time")
        self.therm_setpoint_end_time = raw_data.get("therm_setpoint_end_time")

        self.anticipating = raw_data.get("anticipating")
        self.open_window = raw_data.get("open_window")

        self.cooling_setpoint_temperature = raw_data.get("cooling_setpoint_temperature")
        self.cooling_setpoint_start_time = raw_data.get("cooling_setpoint_start_time")
        self.cooling_setpoint_end_time = raw_data.get("cooling_setpoint_end_time")
        self.cooling_setpoint_mode = raw_data.get("cooling_setpoint_mode")

    async def async_therm_manual(
        self,
        temp: float | None = None,
        end_time: int | None = None,
    ) -> None:
        """Set room temperature set point to manual."""

        await self.async_therm_set(MANUAL, temp, end_time)

    async def async_therm_home(self, end_time: int | None = None) -> None:
        """Set room temperature set point to home."""

        await self.async_therm_set(HOME, end_time=end_time)

    async def async_therm_frostguard(self, end_time: int | None = None) -> None:
        """Set room temperature set point to frostguard."""

        await self.async_therm_set(FROSTGUARD, end_time=end_time)

    async def async_therm_set(
        self,
        mode: str | None = None,
        temp: float | None = None,
        end_time: int | None = None,
        pilot_wire: str | None = None,
    ) -> None:
        """Set room temperature set point."""

        if mode is None:
            mode = MANUAL

        mode = MODE_MAP.get(mode, mode)

        if "NATherm1" in self.device_types or (
            "NRV" in self.device_types
            and not self.home.has_otm()
            and not self.home.has_bns()
        ):
            await self._async_set_thermpoint(mode, temp, end_time)

        else:
            await self._async_therm_set(mode, temp, end_time, pilot_wire)

    async def _async_therm_set(
        self,
        mode: str | None = None,
        temp: float | None = None,
        end_time: int | None = None,
        pilot_wire: str | None = None,
    ) -> bool:
        """Set room temperature set point (OTM)."""
        if pilot_wire is None:
            # in case both are None stop everything
            if mode is None:
                mode = FROSTGUARD
            pilot_wire = climate_setpoint_mode_to_pilot_wire(mode)
            # force back the proper preset mode in case of pilot wire
            # to comply with netatmo model
            if self.support_pilot_wire and self.climate_type == DeviceType.NLC:
                mode = pilot_wire_to_climate_setpoint_mode(pilot_wire)

        if pilot_wire is not None and mode is None:
            mode = MANUAL

        temp_mode_mapping: dict[TemperatureControlMode | None, str] = {
            None: "therm",
            TemperatureControlMode.HEATING: "therm",
            TemperatureControlMode.COOLING: "cooling",
        }

        setpoint_mode_prefix: str = temp_mode_mapping.get(
            self.home.temperature_control_mode,
            "therm",
        )

        room_payload: dict[str, Any] = {
            "id": self.entity_id,
            f"{setpoint_mode_prefix}_setpoint_mode": mode,
        }

        if temp:
            room_payload[f"{setpoint_mode_prefix}_setpoint_temperature"] = temp

        if end_time:
            room_payload[f"{setpoint_mode_prefix}_setpoint_end_time"] = end_time

        if self.support_pilot_wire and pilot_wire:
            room_payload["therm_setpoint_fp"] = pilot_wire

        json_therm_set: dict[str, Any] = {"rooms": [room_payload]}

        return await self.home.async_set_state(json_therm_set)

    async def _async_set_thermpoint(
        self,
        mode: str,
        temp: float | None = None,
        end_time: int | None = None,
    ) -> None:
        """Set room temperature set point (NRV, NATherm1)."""

        post_params: dict[str, str] = {
            "home_id": self.home.entity_id,
            "room_id": self.entity_id,
            "mode": mode,
        }
        # Temp and endtime should only be sent when mode=='manual', but netatmo api can
        # handle that even when mode == 'home' and these settings don't make sense
        if temp is not None:
            post_params["temp"] = str(temp)

        if end_time is not None:
            post_params["endtime"] = str(end_time)

        LOG.debug(
            "Setting room (%s) temperature set point to %s until %s",
            self.entity_id,
            temp,
            end_time,
        )
        await self.home.auth.async_post_api_request(
            endpoint=SETROOMTHERMPOINT_ENDPOINT,
            params=post_params,
        )

    @property
    def boiler_status(self) -> bool | None:
        """Return the boiler status."""

        for module in self.modules.values():
            if hasattr(module, "boiler_status"):
                module = cast("Boiler", module)
                if (boiler_status := module.boiler_status) is not None:
                    return boiler_status

        return None

    @property
    def setpoint_mode(self) -> str:
        """Return the current setpoint mode."""

        return self.therm_setpoint_mode or self.cooling_setpoint_mode or UNKNOWN

    @property
    def setpoint_temperature(self) -> float | None:
        """Return the current setpoint temperature."""

        return (
            self.therm_setpoint_temperature or self.cooling_setpoint_temperature or None
        )

    @property
    def setpoint_fp(self) -> str | None:
        """Return the current setpoint 'Fil pilote (FP)'."""

        return self.therm_setpoint_fp or None

    @property
    def hvac_action(self) -> str:
        """Return the current HVAC action."""

        if self.setpoint_mode == OFF:
            return OFF

        if self.boiler_status is True:
            return HEATING

        if self.heating_power_request is not None and self.heating_power_request > 0:
            return HEATING

        if self.cooling_setpoint_temperature:
            return COOLING

        return IDLE



"""Module to represent Netatmo modules."""

from __future__ import annotations

from dataclasses import dataclass
import logging
from typing import Any

from pyatmo.const import (
    ACCESSORY_GUST_ANGLE_TYPE,
    ACCESSORY_GUST_STRENGTH_TYPE,
    ACCESSORY_RAIN_24H_TYPE,
    ACCESSORY_RAIN_60MIN_TYPE,
    ACCESSORY_RAIN_LIVE_TYPE,
    ACCESSORY_WIND_ANGLE_TYPE,
    ACCESSORY_WIND_STRENGTH_TYPE,
    STATION_HUMIDITY_TYPE,
    STATION_PRESSURE_TYPE,
    STATION_TEMPERATURE_TYPE,
    RawData,
)
from pyatmo.modules.module import (
    BatteryMixin,
    BoilerMixin,
    Camera,
    CO2Mixin,
    DoorTagCategoryMixin,
    FirmwareMixin,
    FloodlightMixin,
    HealthIndexMixin,
    HLSCameraMixin,
    HumidityMixin,
    Module,
    MonitoringMixin,
    NoiseMixin,
    OpenThermMixin,
    PlaceMixin,
    PressureMixin,
    RainMixin,
    RfMixin,
    SirenMixin,
    StatusMixin,
    TemperatureMixin,
    WebRTCCameraMixin,
    WifiMixin,
    WindMixin,
)

LOG: logging.Logger = logging.getLogger(__name__)


class NRV(FirmwareMixin, RfMixin, BatteryMixin, Module):
    """Class to represent a Netatmo NRV."""


class NAC(FirmwareMixin, RfMixin, BatteryMixin, Module):
    """Class to represent a Netatmo NAC."""


class NATherm1(FirmwareMixin, RfMixin, BatteryMixin, BoilerMixin, Module):
    """Class to represent a Netatmo NATherm1."""


class NAPlug(FirmwareMixin, RfMixin, WifiMixin, Module):
    """Class to represent a Netatmo NAPlug."""


class OTH(FirmwareMixin, WifiMixin, BoilerMixin, OpenThermMixin, Module):
    """Class to represent a Netatmo OTH."""


class OTM(FirmwareMixin, RfMixin, BatteryMixin, BoilerMixin, Module):
    """Class to represent a Netatmo OTM."""


class NACamera(HLSCameraMixin, Camera):
    """Class to represent a Netatmo NACamera."""


class NPC(WebRTCCameraMixin, Camera):
    """Class to represent a Netatmo NPC."""


class NOC(SirenMixin, FloodlightMixin, HLSCameraMixin, Camera):
    """Class to represent a Netatmo NOC."""


class NDB(WebRTCCameraMixin, Camera):
    """Class to represent a Netatmo NDB."""


class NAMain(
    TemperatureMixin,
    HumidityMixin,
    CO2Mixin,
    NoiseMixin,
    PressureMixin,
    WifiMixin,
    FirmwareMixin,
    PlaceMixin,
    Module,
):
    """Class to represent a Netatmo NAMain."""


class NAModule1(
    TemperatureMixin,
    HumidityMixin,
    RfMixin,
    FirmwareMixin,
    BatteryMixin,
    PlaceMixin,
    Module,
):
    """Class to represent a Netatmo NAModule1."""


class NAModule2(WindMixin, RfMixin, FirmwareMixin, BatteryMixin, PlaceMixin, Module):
    """Class to represent a Netatmo NAModule2."""


class NAModule3(RainMixin, RfMixin, FirmwareMixin, BatteryMixin, PlaceMixin, Module):
    """Class to represent a Netatmo NAModule3."""


class NAModule4(
    TemperatureMixin,
    CO2Mixin,
    HumidityMixin,
    RfMixin,
    FirmwareMixin,
    BatteryMixin,
    PlaceMixin,
    Module,
):
    """Class to represent a Netatmo NAModule4."""


class NHC(
    TemperatureMixin,
    HumidityMixin,
    CO2Mixin,
    PressureMixin,
    NoiseMixin,
    HealthIndexMixin,
    WifiMixin,
    FirmwareMixin,
    PlaceMixin,
    Module,
):
    """Class to represent a Netatmo NHC."""


class NACamDoorTag(
    StatusMixin, FirmwareMixin, BatteryMixin, RfMixin, DoorTagCategoryMixin, Module
):
    """Class to represent a Netatmo NACamDoorTag."""


class NIS(
    StatusMixin,
    MonitoringMixin,
    FirmwareMixin,
    BatteryMixin,
    RfMixin,
    Module,
):
    """Class to represent a Netatmo NIS."""


class NSD(
    FirmwareMixin,
    Module,
):
    """Class to represent a Netatmo NSD."""


class NCO(
    FirmwareMixin,
    Module,
):
    """Class to represent a Netatmo NCO."""


@dataclass
class Location:
    """Class of Netatmo public weather location."""

    lat_ne: str
    lon_ne: str
    lat_sw: str
    lon_sw: str


class PublicWeatherArea:
    """Class of Netatmo public weather data."""

    location: Location
    required_data_type: str | None
    filtering: bool
    modules: list[dict[str, Any]]

    def __init__(
        self,
        lat_ne: str,
        lon_ne: str,
        lat_sw: str,
        lon_sw: str,
        required_data_type: str | None = None,
        filtering: bool = False,
    ) -> None:
        """Initialize self."""

        self.location = Location(
            lat_ne,
            lon_ne,
            lat_sw,
            lon_sw,
        )
        self.modules = []
        self.required_data_type = required_data_type
        self.filtering = filtering

    def update(self, raw_data: RawData) -> None:
        """Update public weather area with the latest data."""

        self.modules = list(raw_data.get("public", []))

    def stations_in_area(self) -> int:
        """Return available number of stations in area."""

        return len(self.modules)

    def get_latest_rain(self) -> dict[str, Any]:
        """Return latest rain measures."""
        return self.get_accessory_data(ACCESSORY_RAIN_LIVE_TYPE)

    def get_60_min_rain(self) -> dict[str, Any]:
        """Return 60 min rain measures."""
        return self.get_accessory_data(ACCESSORY_RAIN_60MIN_TYPE)

    def get_24_h_rain(self) -> dict[str, Any]:
        """Return 24 h rain measures."""
        return self.get_accessory_data(ACCESSORY_RAIN_24H_TYPE)

    def get_latest_pressures(self) -> dict[str, Any]:
        """Return latest pressure measures."""
        return self.get_latest_station_measures(STATION_PRESSURE_TYPE)

    def get_latest_temperatures(self) -> dict[str, Any]:
        """Return latest temperature measures."""
        return self.get_latest_station_measures(STATION_TEMPERATURE_TYPE)

    def get_latest_humidities(self) -> dict[str, Any]:
        """Return latest humidity measures."""
        return self.get_latest_station_measures(STATION_HUMIDITY_TYPE)

    def get_latest_wind_strengths(self) -> dict[str, Any]:
        """Return latest wind strength measures."""
        return self.get_accessory_data(ACCESSORY_WIND_STRENGTH_TYPE)

    def get_latest_wind_angles(self) -> dict[str, Any]:
        """Return latest wind angle measures."""
        return self.get_accessory_data(ACCESSORY_WIND_ANGLE_TYPE)

    def get_latest_gust_strengths(self) -> dict[str, Any]:
        """Return latest gust strength measures."""
        return self.get_accessory_data(ACCESSORY_GUST_STRENGTH_TYPE)

    def get_latest_gust_angles(self) -> dict[str, Any]:
        """Return latest gust angle measures."""

        return self.get_accessory_data(ACCESSORY_GUST_ANGLE_TYPE)

    def get_latest_station_measures(self, data_type: str) -> dict[str, Any]:
        """Return latest station measures of a given type."""

        measures: dict[str, Any] = {}
        for station in self.modules:
            for module in station["measures"].values():
                if (
                    "type" in module
                    and data_type in module["type"]
                    and "res" in module
                    and module["res"]
                ):
                    measure_index = module["type"].index(data_type)
                    latest_timestamp = sorted(module["res"], reverse=True)[0]
                    measures[station["_id"]] = module["res"][latest_timestamp][
                        measure_index
                    ]

        return measures

    def get_accessory_data(self, data_type: str) -> dict[str, Any]:
        """Return accessory data of a given type."""

        data: dict[str, Any] = {}
        for station in self.modules:
            for module in station["measures"].values():
                if data_type in module:
                    data[station["_id"]] = module[data_type]

        return data



"""Tests for pyatmo.modules.base_class Place and reflection map."""

from pyatmo.modules.base_class import NETATMO_ATTRIBUTES_MAP, Location, Place


def test_place_none_assigns_all_fields() -> None:
    """Place(None) sets every declared field to None without raising."""
    place = Place(None)

    assert place.altitude is None
    assert place.city is None
    assert place.country is None
    assert place.timezone is None
    assert place.location is None


def test_place_none_equality_does_not_raise() -> None:
    """Comparing two Place(None) instances must not raise AttributeError."""
    assert Place(None) == Place(None)


def test_place_dict_without_location() -> None:
    """A dict lacking a valid location keeps other fields, location is None."""
    place = Place({"altitude": 100, "city": "X"})

    assert place.altitude == 100
    assert place.city == "X"
    assert place.country is None
    assert place.timezone is None
    assert place.location is None


def test_place_full_dict() -> None:
    """A fully valid dict populates all fields including a Location."""
    place = Place(
        {
            "altitude": 329,
            "city": "Somewhere",
            "country": "DE",
            "location": [6.1234567, 46.123456],
            "timezone": "Europe/Berlin",
        },
    )

    assert place.altitude == 329
    assert place.city == "Somewhere"
    assert place.country == "DE"
    assert place.timezone == "Europe/Berlin"
    assert place.location == Location(longitude=6.1234567, latitude=46.123456)


def test_place_malformed_location_length() -> None:
    """Wrong-length location leaves location None but keeps other fields."""
    place = Place({"altitude": 100, "city": "X", "location": [1.0]})

    assert place.altitude == 100
    assert place.city == "X"
    assert place.location is None


def test_place_malformed_location_type() -> None:
    """Non-iterable location leaves location None but keeps other fields."""
    place = Place({"altitude": 100, "city": "X", "location": 42})

    assert place.altitude == 100
    assert place.city == "X"
    assert place.location is None


def test_reflection_place_absent_keeps_previous() -> None:
    """When 'place' key is absent, the previous value is kept."""
    fn = NETATMO_ATTRIBUTES_MAP["place"]
    sentinel = object()

    assert fn({}, sentinel) is sentinel


def test_reflection_place_null_keeps_previous() -> None:
    """When 'place' is None, the previous value is kept."""
    fn = NETATMO_ATTRIBUTES_MAP["place"]
    sentinel = object()

    assert fn({"place": None}, sentinel) is sentinel


def test_reflection_place_present_builds_place() -> None:
    """When 'place' is a valid dict, a populated Place is built."""
    fn = NETATMO_ATTRIBUTES_MAP["place"]

    result = fn(
        {
            "place": {
                "altitude": 329,
                "city": "Somewhere",
                "country": "DE",
                "location": [6.1234567, 46.123456],
                "timezone": "Europe/Berlin",
            },
        },
        None,
    )

    assert isinstance(result, Place)
    assert result.altitude == 329
    assert result.location == Location(longitude=6.1234567, latitude=46.123456)


def test_reflection_place_present_invalid_location() -> None:
    """Present 'place' with invalid location builds Place, location None."""
    fn = NETATMO_ATTRIBUTES_MAP["place"]
    sentinel = object()

    result = fn(
        {
            "place": {
                "altitude": 329,
                "city": "Somewhere",
                "location": [1.0, 2.0, 3.0],
            },
        },
        sentinel,
    )

    assert isinstance(result, Place)
    assert result.altitude == 329
    assert result.city == "Somewhere"
    assert result.location is None



"""Module to represent Legrand modules."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from pyatmo.modules.module import (
    BatteryMixin,
    ContactorMixin,
    Dimmer,
    EnergyHistoryLegacyMixin,
    EnergyHistoryMixin,
    Fan,
    FirmwareMixin,
    Module,
    ModuleT,
    OffloadMixin,
    PowerMixin,
    RemoteControlMixin,
    ShutterMixin,
    Switch,
    SwitchMixin,
    WifiMixin,
)

if TYPE_CHECKING:
    from pyatmo.home import Home


LOG: logging.Logger = logging.getLogger(__name__)


class NLG(FirmwareMixin, OffloadMixin, WifiMixin, Module):
    """Legrand gateway."""


class NLT(RemoteControlMixin, BatteryMixin):
    """Legrand global remote control...but also wireless switch, like NLD."""


class NLP(OffloadMixin, Switch):
    """Legrand plug."""


class NLPM(OffloadMixin, Switch):
    """Legrand mobile plug."""


class NLPO(ContactorMixin, OffloadMixin, Switch):
    """Legrand contactor."""


class NLPT(OffloadMixin, Switch):
    """Legrand latching relay/teleruptor."""


class NLPBS(Switch):
    """Legrand british standard plug."""


class NLF(Dimmer):
    """Legrand 2 wire light switch."""


class NLFN(Dimmer):
    """Legrand light switch with neutral."""


class NLFE(Dimmer):
    """Legrand On-Off dimmer switch evolution."""


class NLM(Switch):
    """Legrand light micro module."""


class NLIS(Switch):
    """Legrand double switch."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize NLIS module."""
        super().__init__(home, module)
        if "#" not in self.entity_id:
            # This is a workaround for the fact the this module type has three entries in the API
            # but one of them is not a light switch
            self.device_category = None


class NLD(RemoteControlMixin, BatteryMixin):
    """Legrand Double On/Off dimmer remote. Wireless 2 button switch light."""


class NLL(Switch):
    """Legrand / BTicino italian light switch with neutral."""


class NLV(FirmwareMixin, ShutterMixin, Module):
    """Legrand / BTicino shutters."""


class NLLV(FirmwareMixin, ShutterMixin, Module):
    """Legrand / BTicino shutters."""


class NLLM(FirmwareMixin, ShutterMixin, Module):
    """Legrand / BTicino shutters."""


class NLPC(FirmwareMixin, EnergyHistoryMixin, PowerMixin, Module):
    """Legrand / BTicino connected energy meter."""


class NLE(FirmwareMixin, EnergyHistoryLegacyMixin, Module):
    """Legrand / BTicino connected ecometer. no power supported for the NLE (in the home status API)."""


class NLPS(FirmwareMixin, EnergyHistoryMixin, PowerMixin, Module):
    """Legrand / BTicino smart load shedder."""


class NLC(OffloadMixin, Switch):
    """Legrand / BTicino cable outlet."""


class NLDD(RemoteControlMixin):
    """Legrand NLDD dimmer remote control."""


class NLUP(Switch):
    """Legrand NLUP Power outlet."""


class NLAO(RemoteControlMixin):
    """Legrand wireless batteryless light switch."""


class NLUI(FirmwareMixin, SwitchMixin, Module):
    """Legrand NLUI in-wall switch."""


class NLUF(Dimmer):
    """Legrand NLUF device stub."""


class NLUO(Dimmer):
    """Legrand NLUO device stub."""


class NLLF(Fan, PowerMixin, EnergyHistoryMixin):
    """Legrand NLLF fan/ventilation device."""


class NLunknown(Module):
    """NLunknown device stub."""


class NLAS(RemoteControlMixin):
    """NLAS wireless batteryless scene switch."""


class Z3L(Dimmer):
    """Zigbee 3 Light."""


class Z3V(FirmwareMixin, ShutterMixin, Module):
    """Zigbee 3 roller shutter."""


class EBU(Module):
    """EBU gas meter."""


class NLTS(Module):
    """NLTS motion sensor."""


class NLPD(OffloadMixin, Switch):
    """NLPD dry contact."""


class NLJ(FirmwareMixin, ShutterMixin, Module):
    """Legrand garage door opener."""


class NLY(FirmwareMixin, EnergyHistoryMixin, PowerMixin, Module):
    """Legrand connected 3 phases energy meter."""



"""Collection of helper functions."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, cast

from pyatmo.exceptions import NoDeviceError

if TYPE_CHECKING:
    from pyatmo.const import RawData

LOG: logging.Logger = logging.getLogger(__name__)


def fix_id(raw_data: list[RawData | str]) -> list[RawData | str]:
    """Fix known errors in station ids like superfluous spaces."""

    if not raw_data:
        return raw_data

    for station in raw_data:
        if not isinstance(station, dict):
            continue
        if station.get("_id") is None:
            continue

        station["_id"] = cast("dict", station)["_id"].replace(" ", "")

        for module in station.get("modules", {}):
            module["_id"] = module["_id"].replace(" ", "")

    return raw_data


def extract_raw_data(resp: RawData, tag: str) -> RawData:
    """Extract raw data from server response."""
    if tag == "body":
        return {"public": resp["body"], "errors": []}

    if resp is None or "body" not in resp or tag not in resp["body"]:
        LOG.debug("Server response (tag: %s): %s", tag, resp)
        msg = "No device found, errors in response"
        raise NoDeviceError(msg)

    if tag == "homes":
        homes: list[RawData | str] = fix_id(resp["body"].get(tag))
        if not homes:
            LOG.debug("Server response (tag: %s): %s", tag, resp)
            msg = "No homes found"
            raise NoDeviceError(msg)
        return {
            tag: homes,
            "errors": resp["body"].get("errors", []),
        }

    if not (raw_data := fix_id(resp["body"].get(tag))):
        LOG.debug("Server response (tag: %s): %s", tag, resp)
        msg = "No device data available"
        raise NoDeviceError(msg)

    return {tag: raw_data, "errors": resp["body"].get("errors", [])}



"""Define shared test helpers."""

from __future__ import annotations

import json
import pathlib


def get_fixture_path(filename: str) -> pathlib.Path:
    """Get path of fixture."""
    return pathlib.Path(__file__).parent.joinpath("../fixtures", filename)


def load_fixture(filename):
    """Load a fixture."""
    return get_fixture_path(filename).read_text(encoding="utf-8")


class MockResponse:
    def __init__(self, text, status, headers=None):
        self._text = text
        self.status = status
        self.headers = headers or {}

    async def json(self):
        return self._text

    async def read(self):
        return self._text

    async def __aexit__(self, exc_type, exc, traceback):
        pass

    async def __aenter__(self):
        return self


async def fake_post_request(*_, **kwargs):
    """Return fake data."""
    if "endpoint" not in kwargs:
        return "{}"

    endpoint = kwargs["endpoint"].split("/")[-1]

    if endpoint in "snapshot_720.jpg":
        return b"test stream image bytes"

    if endpoint in [
        "setpersonsaway",
        "setpersonshome",
        "setstate",
        "setroomthermpoint",
        "setthermmode",
        "switchhomeschedule",
    ]:
        payload = f'{{"{endpoint}": true}}'

    elif endpoint == "homestatus":
        home_id = kwargs.get("params", {}).get("home_id")
        payload = json.loads(load_fixture(f"{endpoint}_{home_id}.json"))

    elif endpoint == "getmeasure":
        module_id = kwargs.get("params", {}).get("module_id", "")
        measure_type = kwargs.get("params", {}).get("type", "")
        payload = json.loads(
            load_fixture(
                f"{endpoint}_{measure_type}_{module_id.replace(':', '_')}.json",
            ),
        )

    else:
        postfix = kwargs.get("POSTFIX")
        if postfix is not None:
            payload = json.loads(load_fixture(f"{endpoint}_{postfix}.json"))
        else:
            payload = json.loads(load_fixture(f"{endpoint}.json"))

    return MockResponse(payload, 200)


async def fake_post_request_multi(*args, **kwargs):
    kwargs["POSTFIX"] = "multi"
    return await fake_post_request(*args, **kwargs)



import asyncio

from aiohttp import ClientSession

import pyatmo
from pyatmo.auth import AbstractAsyncAuth
from pyatmo.modules.module import MeasureInterval

MY_TOKEN_FROM_NETATMO = "MY_TOKEN"


class MyAuth(AbstractAsyncAuth):
    async def async_get_access_token(self):
        return MY_TOKEN_FROM_NETATMO


async def main():
    session = ClientSession()
    async_auth = MyAuth(session)
    account = pyatmo.AsyncAccount(async_auth)

    t = asyncio.create_task(account.async_update_topology())
    home_id = "MY_HOME_ID"
    module_id = "MY_MODULE_ID"

    await asyncio.gather(t)

    await account.async_update_status(home_id=home_id)

    strt = 1709766000 + 10 * 60
    end = 1709852400 + 10 * 60
    await account.async_update_measures(
        home_id=home_id,
        module_id=module_id,
        interval=MeasureInterval.HALF_HOUR,
        start_time=strt,
        end_time=end,
    )


if __name__ == "__main__":
    topology = asyncio.run(main())



"""Module to represent a Netatmo module."""

from __future__ import annotations

from datetime import UTC, datetime, timedelta
from enum import Enum
import logging
from operator import itemgetter
from time import time
from typing import TYPE_CHECKING, Any

from aiohttp import ClientConnectorError, ClientResponse

from pyatmo.const import (
    GETMEASURE_ENDPOINT,
    SETSTATE_ENDPOINT,
    WEBRTC_OFFER_ENDPOINT,
    WEBRTC_TERMINATE_ENDPOINT,
    RawData,
)
from pyatmo.exceptions import ApiError
from pyatmo.modules.base_class import (
    EntityBase,
    NetatmoBase,
    Place,
    bridged_module_ids,
    update_name,
)
from pyatmo.modules.device_types import (
    DEVICE_CATEGORY_MAP,
    ApplianceType,
    BoilerControl,
    BoilerError,
    DeviceCategory,
    DeviceType,
    DhwControl,
    DoorTagCategory,
)
from pyatmo.webrtc import WebRTCAnswer, WebRTCStream

if TYPE_CHECKING:
    from pyatmo.event import Event
    from pyatmo.home import Home


LOG = logging.getLogger(__name__)


ModuleT = dict[str, Any]
# Hide from features list
ATTRIBUTE_FILTER = {
    "battery_state",
    "battery_level",
    "battery_percent",
    "date_min_temp",
    "date_max_temp",
    "name",
    "entity_id",
    "device_id",
    "modules",
    "firmware_revision",
    "firmware_name",
    "home",
    "bridge",
    "room_id",
    "device_category",
    "device_type",
    "features",
    "history_features",
    "history_features_values",
    "appliance_type",
    "doortag_category",
    "last_seen",
    "setup_date",
    "boiler_control",
    "boiler_error",
    "dhw_control",
    "error_code",
}


def process_battery_state(data: str) -> int:
    """Process battery data and return percent (int) for display."""

    mapping = {
        "max": 100,
        "full": 90,
        "high": 75,
        "medium": 50,
        "low": 25,
        "very_low": 10,
    }
    return mapping.get(data, 0)


class FirmwareMixin(EntityBase):
    """Mixin for firmware data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize firmware mixin."""
        super().__init__(home, module)
        self.firmware_revision: int | None = None
        self.firmware_name: str | None = None


class WifiMixin(EntityBase):
    """Mixin for wifi data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize wifi mixin."""
        super().__init__(home, module)
        self.wifi_strength: int | None = None


class RfMixin(EntityBase):
    """Mixin for rf data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize rf mixin."""

        super().__init__(home, module)
        self.rf_strength: int | None = None


class RainMixin(EntityBase):
    """Mixin for rain data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize rain mixin."""

        super().__init__(home, module)
        self.rain: float | None = None
        self.sum_rain_1: float | None = None
        self.sum_rain_24: float | None = None


class WindMixin(EntityBase):
    """Mixin for wind data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize wind mixin."""

        super().__init__(home, module)
        self.wind_strength: int | None = None
        self.wind_angle: int | None = None
        self.gust_strength: int | None = None
        self.gust_angle: int | None = None

    @property
    def wind_direction(self) -> str | None:
        """Return wind direction."""

        return None if self.wind_angle is None else process_angle(self.wind_angle)

    @property
    def gust_direction(self) -> str | None:
        """Return gust direction."""

        return None if self.gust_angle is None else process_angle(self.gust_angle)


def process_angle(angle: int) -> str:
    """Process angle and return string for display."""
    angle_mapping = {
        (0, 30): "N",
        (30, 60): "NE",
        (60, 120): "E",
        (120, 150): "SE",
        (150, 210): "S",
        (210, 240): "SW",
        (240, 300): "W",
        (300, 330): "NW",
        (330, 360): "N",
    }
    for (lower, upper), direction in angle_mapping.items():
        if lower <= angle < upper:
            return direction
    return "N"  # Default case


class TemperatureMixin(EntityBase):
    """Mixin for temperature data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize temperature mixin."""

        super().__init__(home, module)
        self.temperature: float | None = None
        self.temp_min: float | None = None
        self.temp_max: float | None = None
        self.temp_trend: str | None = None
        self.min_temp: float | None = None
        self.max_temp: float | None = None
        self.date_min_temp: int | None = None
        self.date_max_temp: int | None = None


class HumidityMixin(EntityBase):
    """Mixin for humidity data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize humidity mixin."""

        super().__init__(home, module)
        self.humidity: int | None = None


class CO2Mixin(EntityBase):
    """Mixin for CO2 data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize CO2 mixin."""

        super().__init__(home, module)
        self.co2: int | None = None


class HealthIndexMixin(EntityBase):
    """Mixin for health index data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize health index mixin."""

        super().__init__(home, module)
        self.health_idx: int | None = None


class NoiseMixin(EntityBase):
    """Mixin for noise data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize noise mixin."""

        super().__init__(home, module)
        self.noise: int | None = None


class PressureMixin(EntityBase):
    """Mixin for pressure data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize pressure mixin."""

        super().__init__(home, module)
        self.pressure: float | None = None
        self.absolute_pressure: float | None = None
        self.pressure_trend: str | None = None


class BoilerMixin(EntityBase):
    """Mixin for boiler data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize boiler mixin."""

        super().__init__(home, module)
        self.boiler_status: bool | None = None
        self.boiler_valve_comfort_boost: bool | None = None


class OpenThermMixin(EntityBase):
    """Mixin for OpenTherm boiler diagnostics (OTH)."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize OpenTherm mixin."""

        super().__init__(home, module)
        self.boiler_control: BoilerControl | None = None
        self.boiler_error: BoilerError | None = None
        self.dhw_control: DhwControl | None = None


class CoolerMixin(EntityBase):
    """Mixin for cooler data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize cooler mixin."""

        super().__init__(home, module)
        self.cooler_status: bool | None = None


class BatteryMixin(EntityBase):
    """Mixin for battery data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize battery mixin."""

        super().__init__(home, module)
        self.battery_state: str | None = None
        self.battery_level: int | None = None
        self.battery_percent: int | None = None

    @property
    def battery(self) -> int:
        """Return battery percent."""

        if self.battery_percent is not None:
            return self.battery_percent
        if self.battery_state is None:
            return 0
        return process_battery_state(self.battery_state)


class PlaceMixin(EntityBase):
    """Mixin for place data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize place mixin."""

        super().__init__(home, module)
        self.place: Place | None = None


class DimmableMixin(EntityBase):
    """Mixin for dimmable data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize dimmable mixin."""

        super().__init__(home, module)
        self.brightness: int | None = None

    async def async_set_brightness(self, brightness: int) -> bool:
        """Set brightness."""

        json_brightness = {
            "modules": [
                {
                    "id": self.entity_id,
                    "brightness": max(min(100, brightness), -1),
                    "bridge": self.bridge,
                },
            ],
        }
        return await self.home.async_set_state(json_brightness)


class ApplianceTypeMixin(EntityBase):
    """Mixin for appliance type data."""

    device_category: DeviceCategory | None
    appliance_type: ApplianceType | None

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize appliance type mixin."""

        super().__init__(home, module)
        self.appliance_type: ApplianceType | None = module.get(
            "appliance_type",
            ApplianceType.unknown,
        )


class DoorTagCategoryMixin(EntityBase):
    """Mixin for category data."""

    doortag_category: DoorTagCategory | None

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize category mixin."""

        super().__init__(home, module)
        self.doortag_category: DoorTagCategory | None = module.get(
            "category",
            DoorTagCategory.unknown,
        )


class PowerMixin(EntityBase):
    """Mixin for power data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize power mixin."""

        super().__init__(home, module)
        self.power: int | None = None
        self.history_features.add("power")


class EventMixin(EntityBase):
    """Mixin for event data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize event mixin."""

        super().__init__(home, module)
        self.events: list[Event] = []


class ContactorMixin(EntityBase):
    """Mixin for contactor data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize contactor mixin."""

        super().__init__(home, module)
        self.contactor_mode: str | None = None


class OffloadMixin(EntityBase):
    """Mixin for offload data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize offload mixin."""

        super().__init__(home, module)
        self.offload: bool | None = None
        self.offload_meters: list[str] | None = None


class SwitchMixin(EntityBase):
    """Mixin for switch data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize switch mixin."""

        super().__init__(home, module)
        self.on: bool | None = None

    async def async_set_switch(self, target_position: bool) -> bool:
        """Set switch to target position."""

        json_switch = {
            "modules": [
                {
                    "id": self.entity_id,
                    "on": target_position,
                    "bridge": self.bridge,
                },
            ],
        }
        return await self.home.async_set_state(json_switch)

    async def async_on(self) -> bool:
        """Switch on."""

        return await self.async_set_switch(True)

    async def async_off(self) -> bool:
        """Switch off."""

        return await self.async_set_switch(False)


class FanSpeedMixin(EntityBase):
    """Mixin for fan speed data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize fan speed mixin."""

        super().__init__(home, module)
        self.fan_speed: int | None = None

    async def async_set_fan_speed(self, speed: int) -> bool:
        """Set fan speed."""

        json_fan_speed = {
            "modules": [
                {
                    "id": self.entity_id,
                    # fan speed is clamped between 1 and 2
                    # since only NLLF is such a device
                    # and it can only supports fan_speed 1 or 2
                    "fan_speed": max(min(2, speed), 1),
                    "bridge": self.bridge,
                },
            ],
        }
        return await self.home.async_set_state(json_fan_speed)


class ShutterMixin(EntityBase):
    """Mixin for shutter data."""

    __open_position = 100
    __close_position = 0
    __stop_position = -1
    __preferred_position = -2

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize shutter mixin."""

        super().__init__(home, module)
        self.current_position: int | None = None
        self.target_position: int | None = None
        self.target_position__step: int | None = None

    async def async_set_target_position(self, target_position: int) -> bool:
        """Set shutter to target position."""

        # in case of a too low value, we default to stop and not the preferred position
        # We check against __preferred_position that is the lower known value
        if target_position < self.__preferred_position:
            target_position = self.__stop_position

        json_roller_shutter = {
            "modules": [
                {
                    "id": self.entity_id,
                    "target_position": min(self.__open_position, target_position),
                    "bridge": self.bridge,
                },
            ],
        }
        return await self.home.async_set_state(json_roller_shutter)

    async def async_open(self) -> bool:
        """Open shutter."""

        return await self.async_set_target_position(self.__open_position)

    async def async_close(self) -> bool:
        """Close shutter."""

        return await self.async_set_target_position(self.__close_position)

    async def async_stop(self) -> bool:
        """Stop shutter."""

        return await self.async_set_target_position(self.__stop_position)

    async def async_move_to_preferred_position(self) -> bool:
        """Move shutter to preferred position."""

        return await self.async_set_target_position(self.__preferred_position)


class CameraMixinBase(EntityBase):
    """Base class for camera mixins."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize camera mixin."""

        super().__init__(home, module)
        self.sd_status: int | None = None
        self.vpn_url: str | None = None
        self.local_url: str | None = None
        self.is_local: bool | None = None
        self.alim_status: int | None = None
        self.device_type: DeviceType

        self._force_vpn_url = False

    @property
    def camera_url(self) -> str | None:
        """Return the camera URL, if available.

        Depending on the camera streaming protocol, this URL can be used to retrieve:
          - The live snapshot (HLS and WebRTC).
          - The live stream (HLS only).
        """
        return self.local_url or self.vpn_url

    async def async_get_live_snapshot(self) -> bytes | None:
        """Fetch live camera image."""

        url = self.camera_url

        if not url:
            return None

        resp = await self.home.auth.async_get_image(
            base_url=f"{url}",
            endpoint="/live/snapshot_720.jpg",
        )

        return resp if isinstance(resp, bytes) else None

    async def async_update_camera_urls(self) -> None:
        """Update and validate the camera urls."""

        if self._force_vpn_url:
            self.local_url = None
            return

        if self.vpn_url and self.is_local:
            temp_local_url = await self._async_check_url(self.vpn_url)
            if temp_local_url:
                try:
                    self.local_url = await self._async_check_url(
                        temp_local_url,
                    )
                except (TimeoutError, ClientConnectorError) as exc:
                    LOG.debug("Cannot connect to %s - reason: %s", temp_local_url, exc)
                    self.is_local = False
                    self.local_url = None

    async def _async_check_url(self, url: str) -> str | None:
        """Validate camera url."""

        try:
            resp = await self.home.auth.async_post_api_request(
                base_url=f"{url}",
                endpoint="/command/ping",
            )

        except ApiError:
            LOG.debug("Api error for camera url %s", url)
            return None

        if isinstance(resp, bytes):
            msg = "Invalid response from camera url"
            raise ApiError(msg)

        resp_data = await resp.json()
        return resp_data.get("local_url") if resp_data else None


class HLSCameraMixin(CameraMixinBase):
    """Mixin for cameras using the HLS protocol."""


class WebRTCCameraMixin(CameraMixinBase):
    """Mixin for cameras using the WebRTC protocol."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize WebRTC camera mixin."""

        super().__init__(home, module)

        # WebRTC cameras don't support accessing the live stream nor snapshot via the local URL
        self._force_vpn_url = True

    async def async_start_stream(self, session_id: str, sdp_offer: str) -> WebRTCAnswer:
        """Start WebRTC streaming session.

        Netatmo cameras do not support ICE trickle when accessed through third-party applications,
        so the SDP offer must include all ICE candidates.
        """

        params = {
            "home_id": self.home.entity_id,
            "device_id": self.entity_id,
            "sdp": sdp_offer,
            "session_id": session_id,
        }

        resp = await self.home.auth.async_post_api_request(
            endpoint=WEBRTC_OFFER_ENDPOINT, params=params
        )

        json_resp = await resp.json()
        resp_body = self._parse_webrtc_response_body(json_resp)

        try:
            if session_id != resp_body["session_id"]:
                msg = "Invalid WebRTC answer: session ID mismatch"
                raise ApiError(msg)

            tag_id = resp_body["tag_id"]
            sdp_answer = resp_body["sdpAnswer"]
        except KeyError as exc:
            msg = f"Invalid WebRTC answer: missing {exc} field"
            raise ApiError(msg) from exc

        return WebRTCAnswer(WebRTCStream(session_id, tag_id), sdp_answer)

    async def async_stop_stream(self, stream: WebRTCStream) -> None:
        """Stop an active WebRTC streaming session."""

        params = {
            "home_id": self.home.entity_id,
            "device_id": self.entity_id,
            "session_id": stream.session_id,
            "tag_id": stream.tag_id,
        }

        resp = await self.home.auth.async_post_api_request(
            endpoint=WEBRTC_TERMINATE_ENDPOINT, params=params
        )

        json_resp = await resp.json()
        self._parse_webrtc_response_body(json_resp)

    def _parse_webrtc_response_body(self, json_resp: dict[str, Any]) -> dict[str, Any]:
        """Extract the body from a WebRTC API response and check for errors."""

        if not json_resp:
            msg = "Empty WebRTC response"
            raise ApiError(msg)

        try:
            resp_body = json_resp["body"]

            if resp_body["status"] != "ok":
                error_dict = resp_body["error"]
                error_code = error_dict["code"]
                error_msg = error_dict["message"]

                msg = f"WebRTC API error: {error_msg} ({error_code})"
                raise ApiError(msg)
        except KeyError as exc:
            msg = f"Invalid WebRTC response: missing {exc} field"
            raise ApiError(msg) from exc

        return resp_body


class FloodlightMixin(EntityBase):
    """Mixin for floodlight data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize floodlight mixin."""

        super().__init__(home, module)
        self.floodlight: str | None = None

    async def async_set_floodlight_state(self, state: str) -> bool:
        """Set floodlight state."""

        json_floodlight_state = {
            "modules": [
                {
                    "id": self.entity_id,
                    "floodlight": state,
                },
            ],
        }
        return await self.home.async_set_state(json_floodlight_state)

    async def async_floodlight_on(self) -> bool:
        """Turn on floodlight."""

        return await self.async_set_floodlight_state("on")

    async def async_floodlight_off(self) -> bool:
        """Turn off floodlight."""

        return await self.async_set_floodlight_state("off")

    async def async_floodlight_auto(self) -> bool:
        """Set floodlight to auto mode."""

        return await self.async_set_floodlight_state("auto")


class SirenMixin(EntityBase):
    """Mixin for siren data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize siren mixin."""

        super().__init__(home, module)
        self.siren_status: str | None = None

    async def async_set_siren_state(
        self, state: str, base_url: str | None = None
    ) -> bool:
        """Set siren state.

        Uses the public OAuth2 API by default. Pass base_url=SIREN_BASE_URL
        to route via app.netatmo.net, which currently accepts siren_status
        where the public API rejects it (error code 21).
        """

        resp = await self.home.auth.async_post_api_request(
            endpoint=SETSTATE_ENDPOINT,
            base_url=base_url,
            params={
                "json": {
                    "home": {
                        "id": self.home.entity_id,
                        "modules": [
                            {
                                "id": self.entity_id,
                                "siren_status": state,
                            },
                        ],
                    },
                },
            },
        )
        return (await resp.json()).get("status") == "ok"

    async def async_siren_on(self, base_url: str | None = None) -> bool:
        """Turn on siren."""

        return await self.async_set_siren_state("sound", base_url=base_url)

    async def async_siren_off(self, base_url: str | None = None) -> bool:
        """Turn off siren."""

        return await self.async_set_siren_state("no_sound", base_url=base_url)


class StatusMixin(EntityBase):
    """Mixin for status data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize status mixin."""

        super().__init__(home, module)
        self.status: str | None = None


class MonitoringMixin(EntityBase):
    """Mixin for monitoring data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize monitoring mixin."""

        super().__init__(home, module)
        self.monitoring: bool | None = None

    async def async_set_monitoring_state(self, state: str) -> bool:
        """Set monitoring state."""

        json_monitoring_state = {
            "modules": [
                {
                    "id": self.entity_id,
                    "monitoring": state,
                },
            ],
        }
        return await self.home.async_set_state(json_monitoring_state)

    async def async_monitoring_on(self) -> bool:
        """Turn on monitoring."""

        return await self.async_set_monitoring_state("on")

    async def async_monitoring_off(self) -> bool:
        """Turn off monitoring."""

        return await self.async_set_monitoring_state("off")


class MeasureInterval(Enum):
    """Measure interval."""

    HALF_HOUR = "30min"
    HOUR = "1hour"
    THREE_HOURS = "3hours"
    DAY = "1day"
    WEEK = "1week"
    MONTH = "1month"


class MeasureType(Enum):
    """Measure type."""

    BOILERON = "boileron"
    BOILEROFF = "boileroff"
    SUM_BOILER_ON = "sum_boiler_on"
    SUM_BOILER_OFF = "sum_boiler_off"
    SUM_ENERGY_ELEC = "sum_energy_buy_from_grid"
    SUM_ENERGY_ELEC_BASIC = "sum_energy_buy_from_grid$0"
    SUM_ENERGY_ELEC_PEAK = "sum_energy_buy_from_grid$1"
    SUM_ENERGY_ELEC_OFF_PEAK = "sum_energy_buy_from_grid$2"
    SUM_ENERGY_PRICE = "sum_energy_buy_from_grid_price"
    SUM_ENERGY_PRICE_BASIC = "sum_energy_buy_from_grid_price$0"
    SUM_ENERGY_PRICE_PEAK = "sum_energy_buy_from_grid_price$1"
    SUM_ENERGY_PRICE_OFF_PEAK = "sum_energy_buy_from_grid_price$2"
    SUM_ENERGY_ELEC_LEGACY = "sum_energy_elec"
    SUM_ENERGY_ELEC_BASIC_LEGACY = "sum_energy_elec$0"
    SUM_ENERGY_ELEC_PEAK_LEGACY = "sum_energy_elec$1"
    SUM_ENERGY_ELEC_OFF_PEAK_LEGACY = "sum_energy_elec$2"


MEASURE_INTERVAL_TO_SECONDS = {
    MeasureInterval.HALF_HOUR: 1800,
    MeasureInterval.HOUR: 3600,
    MeasureInterval.THREE_HOURS: 10800,
    MeasureInterval.DAY: 86400,
    MeasureInterval.WEEK: 604800,
    MeasureInterval.MONTH: 2592000,
}

ENERGY_FILTERS: str = (
    f"{MeasureType.SUM_ENERGY_ELEC.value},"
    f"{MeasureType.SUM_ENERGY_ELEC_BASIC.value},"
    f"{MeasureType.SUM_ENERGY_ELEC_PEAK.value},"
    f"{MeasureType.SUM_ENERGY_ELEC_OFF_PEAK.value}"
)
ENERGY_FILTERS_LEGACY: str = (
    f"{MeasureType.SUM_ENERGY_ELEC_LEGACY.value},"
    f"{MeasureType.SUM_ENERGY_ELEC_BASIC_LEGACY.value},"
    f"{MeasureType.SUM_ENERGY_ELEC_PEAK_LEGACY.value},"
    f"{MeasureType.SUM_ENERGY_ELEC_OFF_PEAK_LEGACY.value}"
)
ENERGY_FILTERS_MODES: list[str] = ["generic", "basic", "peak", "off_peak"]


def compute_riemann_sum(
    power_data: list[tuple[int, float]],
    conservative: bool = False,
) -> float:
    """Compute energy from power with a rieman sum."""

    delta_energy = 0.0
    if power_data and len(power_data) > 1:
        # compute a rieman sum, as best as possible , trapezoidal, taking pessimistic asumption
        # as we don't want to artifically go up the previous one
        # (except in rare exceptions like reset, 0 , etc)

        for i in range(len(power_data) - 1):
            dt_h = float(power_data[i + 1][0] - power_data[i][0]) / 3600.0

            if conservative:
                d_p_w = 0.0
            else:
                d_p_w = abs(float(power_data[i + 1][1] - power_data[i][1]))

            d_nrj_wh = dt_h * (
                min(power_data[i + 1][1], power_data[i][1]) + 0.5 * d_p_w
            )

            delta_energy += d_nrj_wh

    return delta_energy


class EnergyHistoryMixin(EntityBase):
    """Mixin for Energy history data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize history mixin."""

        super().__init__(home, module)
        self.historical_data: list[dict[str, Any]] | None = None
        self.start_time: float | None = None
        self.end_time: float | None = None
        self.interval: MeasureInterval | None = None
        self.sum_energy_elec: float = 0.0
        self.sum_energy_elec_peak: float = 0.0
        self.sum_energy_elec_off_peak: float = 0.0
        self._anchor_for_power_adjustment: float | None = None
        self.in_reset: bool = False

    def reset_measures(
        self,
        start_power_time: datetime | None,
        in_reset: bool = True,
    ) -> None:
        """Reset energy measures."""

        self.in_reset = in_reset
        self.historical_data = []
        self.sum_energy_elec = 0.0
        self.sum_energy_elec_peak = 0.0
        self.sum_energy_elec_off_peak = 0.0
        if start_power_time is None:
            self._anchor_for_power_adjustment = start_power_time
        else:
            self._anchor_for_power_adjustment = int(start_power_time.timestamp())

    def get_sum_energy_elec_power_adapted(
        self,
        to_ts: float | None = None,
        conservative: bool = False,
    ) -> tuple[float, float]:
        """Compute proper energy value with adaptation from power."""

        v = self.sum_energy_elec

        delta_energy = 0.0

        if not self.in_reset:
            if to_ts is None:
                to_ts = int(time())

            from_ts = self._anchor_for_power_adjustment

            if (
                from_ts is not None
                and from_ts < to_ts
                and isinstance(self, PowerMixin)
                and isinstance(self, NetatmoBase)
            ):
                power_data = self.get_history_data(
                    "power",
                    from_ts=from_ts,
                    to_ts=to_ts,
                )
                if isinstance(
                    self,
                    EnergyHistoryMixin,
                ):  # well to please the linter....
                    delta_energy = compute_riemann_sum(power_data, conservative)

        return v, delta_energy

    def _log_energy_error(
        self,
        start_time: float,
        end_time: float,
        msg: str | None = None,
        body: dict | None = None,
    ) -> None:
        LOG.debug(
            "ENERGY collection error %s %s %s %s %s %s %s",
            msg,
            self.name,
            datetime.fromtimestamp(start_time),
            datetime.fromtimestamp(end_time),
            start_time,
            end_time,
            body or "NO BODY",
        )

    async def async_update_measures(
        self,
        start_time: int | None = None,
        end_time: int | None = None,
        interval: MeasureInterval = MeasureInterval.HOUR,
        days: int = 7,
    ) -> None:
        """Update historical data."""

        if end_time is None:
            end_time = int(datetime.now().timestamp())

        if start_time is None:
            end = datetime.fromtimestamp(end_time)
            start_time = int((end - timedelta(days=days)).timestamp())

        prev_start_time = self.start_time
        prev_end_time = self.end_time

        self.start_time = start_time
        self.end_time = end_time

        # the legrand/netatmo handling of start and endtime is very peculiar
        # for 30mn/1h/3h intervals : in fact the starts is asked_start + intervals/2 !
        # => so shift of 15mn, 30mn and 1h30
        # for 1day : start is ALWAYS 12am (half day) of the first day of the range
        # for 1week : it will be half week ALWAYS, ie on a thursday at 12am (half day)
        # in fact in the case for all intervals the reported dates are "the middle" of the ranges

        delta_range = MEASURE_INTERVAL_TO_SECONDS.get(interval, 0) // 2

        _, raw_data = await self._energy_api_calls(start_time, end_time, interval)

        hist_good_vals = await self._get_aligned_energy_values_and_mode(
            start_time,
            end_time,
            delta_range,
            raw_data,
        )

        prev_sum_energy_elec = self.sum_energy_elec
        self.sum_energy_elec = 0.0
        self.sum_energy_elec_peak = 0.0
        self.sum_energy_elec_off_peak = 0.0

        # no data at all: we know nothing for the end: best guess, it is the start
        self._anchor_for_power_adjustment = start_time

        self.in_reset = False

        if len(hist_good_vals) == 0:
            # nothing has been updated or changed it can nearly be seen as an error, but the api is answering correctly
            # so we probably have to reset to 0 anyway as it means there were no exisitng
            # historical data for this time range

            LOG.debug(
                "NO VALUES energy update %s from: %s to %s,  prev_sum=%s",
                self.name,
                datetime.fromtimestamp(start_time),
                datetime.fromtimestamp(end_time),
                prev_sum_energy_elec if prev_sum_energy_elec is not None else "NOTHING",
            )
        else:
            await self._prepare_exported_historical_data(
                start_time,
                end_time,
                delta_range,
                hist_good_vals,
                prev_end_time or 0.0,
                prev_start_time or 0.0,
                prev_sum_energy_elec or 0.0,
            )

    async def _prepare_exported_historical_data(
        self,
        start_time: float,
        end_time: float,
        delta_range: float,
        hist_good_vals: list[tuple[int, float, list[float]]],
        prev_end_time: float,
        prev_start_time: float,
        prev_sum_energy_elec: float | None,
    ) -> None:
        self.historical_data = []
        computed_start = 0.0
        computed_end = 0.0
        computed_end_for_calculus = 0.0
        for cur_start_time, val, vals in hist_good_vals:
            self.sum_energy_elec += val

            modes = []
            val_modes = []

            for i, v in enumerate(vals):
                if v is not None:
                    modes.append(ENERGY_FILTERS_MODES[i])
                    val_modes.append(v)
                    if ENERGY_FILTERS_MODES[i] == "off_peak":
                        self.sum_energy_elec_off_peak += v
                    elif ENERGY_FILTERS_MODES[i] == "peak":
                        self.sum_energy_elec_peak += v

            c_start = cur_start_time
            c_end = cur_start_time + 2 * delta_range

            if computed_start == 0:
                computed_start = c_start
            computed_end = c_end

            computed_end_for_calculus = c_end

            start_time_string = f"{datetime.fromtimestamp(c_start + 1, tz=UTC).isoformat().split('+')[0]}Z"
            end_time_string = (
                f"{datetime.fromtimestamp(c_end, tz=UTC).isoformat().split('+')[0]}Z"
            )
            self.historical_data.append(
                {
                    "duration": (2 * delta_range) // 60,
                    "startTime": start_time_string,
                    "endTime": end_time_string,
                    "Wh": val,
                    "energyMode": modes,
                    "WhPerModes": val_modes,
                    "startTimeUnix": c_start,
                    "endTimeUnix": c_end,
                },
            )
        if (
            prev_sum_energy_elec is not None
            and prev_sum_energy_elec > self.sum_energy_elec
        ):
            msg = (
                "ENERGY GOING DOWN %s from: %s to %s "
                "computed_start: %s, computed_end: %s, "
                "sum=%f prev_sum=%f prev_start: %s, prev_end %s"
            )
            LOG.debug(
                msg,
                self.name,
                datetime.fromtimestamp(start_time),
                datetime.fromtimestamp(end_time),
                datetime.fromtimestamp(computed_start),
                datetime.fromtimestamp(computed_end),
                self.sum_energy_elec,
                prev_sum_energy_elec,
                datetime.fromtimestamp(prev_start_time),
                datetime.fromtimestamp(prev_end_time),
            )
        else:
            msg = (
                "Success in energy update %s from: %s to %s "
                "computed_start: %s, computed_end: %s , sum=%s prev_sum=%s"
            )
            LOG.debug(
                msg,
                self.name,
                datetime.fromtimestamp(start_time),
                datetime.fromtimestamp(end_time),
                datetime.fromtimestamp(computed_start),
                datetime.fromtimestamp(computed_end),
                self.sum_energy_elec,
                prev_sum_energy_elec if prev_sum_energy_elec is not None else "NOTHING",
            )

        self._anchor_for_power_adjustment = computed_end_for_calculus

    async def _get_aligned_energy_values_and_mode(
        self,
        start_time: float,
        end_time: float,
        delta_range: float,
        raw_data: dict,
    ) -> list[Any]:
        hist_good_vals = []
        values_lots = raw_data

        for values_lot in values_lots:
            try:
                start_lot_time = int(values_lot["beg_time"])
            except KeyError:
                self._log_energy_error(
                    start_time,
                    end_time,
                    msg="beg_time missing",
                    body=values_lots,
                )
                msg = (
                    f"Energy badly formed resp beg_time missing: {values_lots} - "
                    f"module: {self.name}"
                )
                raise ApiError(
                    msg,
                ) from None

            interval_sec = values_lot.get("step_time")
            if interval_sec is None:
                if len(values_lot.get("value", [])) > 1:
                    self._log_energy_error(
                        start_time,
                        end_time,
                        msg="step_time missing",
                        body=values_lots,
                    )
                interval_sec = 2 * delta_range
            else:
                interval_sec = int(interval_sec)

            # align the start on the begining of the segment
            cur_start_time = start_lot_time - interval_sec // 2
            for val_arr in values_lot.get("value", []):
                vals: list[int | None] = []
                val = 0
                for v in val_arr:
                    if v is not None:
                        v = int(v)
                        val += v
                        vals.append(v)
                    else:
                        vals.append(None)

                hist_good_vals.append((cur_start_time, val, vals))
                cur_start_time = cur_start_time + interval_sec

        return sorted(hist_good_vals, key=itemgetter(0))

    def _get_energy_filers(self) -> str:
        return ENERGY_FILTERS

    async def _energy_api_calls(
        self,
        start_time: float,
        end_time: float,
        interval: MeasureInterval,
    ) -> tuple[str, Any]:
        filters: str = self._get_energy_filers()

        params = {
            "device_id": self.bridge,
            "module_id": self.entity_id,
            "scale": interval.value,
            "type": filters,
            "date_begin": start_time,
            "date_end": end_time,
        }

        resp: ClientResponse = await self.home.auth.async_post_api_request(
            endpoint=GETMEASURE_ENDPOINT,
            params=params,
        )

        rw_dt_f = await resp.json()
        rw_dt = rw_dt_f.get("body")

        if rw_dt is None:
            self._log_energy_error(
                start_time,
                end_time,
                msg=f"direct from {filters}",
                body=rw_dt_f,
            )
            msg: str = (
                f"Energy badly formed resp: {rw_dt_f} - "
                f"module: {self.name} - "
                f"when accessing '{filters}'"
            )
            raise ApiError(msg)

        raw_data = rw_dt

        return filters, raw_data


class EnergyHistoryLegacyMixin(EnergyHistoryMixin):
    """Mixin for Energy history data, Using legacy APis (used for NLE)."""

    def _get_energy_filers(self) -> str:
        return ENERGY_FILTERS_LEGACY


class Module(NetatmoBase):
    """Class to represent a Netatmo module."""

    device_type: DeviceType
    device_category: DeviceCategory | None
    room_id: str | None

    modules: list[str] | None
    reachable: bool | None
    last_seen: int | None
    setup_date: int | None
    error_code: int | None
    features: set[str]

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize a Netatmo module instance."""

        super().__init__(module)

        self.device_type = DeviceType(module["type"])

        self.home = home
        self.room_id = module.get("room_id")
        self.reachable = module.get("reachable")
        self.last_seen = module.get("last_seen")
        self.setup_date = module.get("setup_date")
        self.error_code = None
        self.bridge = module.get("bridge")
        self.modules = bridged_module_ids(module)
        self.device_category = DEVICE_CATEGORY_MAP.get(self.device_type)
        self.features = set()

    async def update(self, raw_data: RawData) -> None:
        """Update module with the latest data."""

        self.update_topology(raw_data)

        if (
            self.bridge
            and self.bridge in self.home.modules
            and hasattr(self, "device_category")
            and self.device_category == "weather"
        ):
            self.name = update_name(self.name, self.home.modules[self.bridge].name)

        self.update_features()

        # If we have an NLE as a bridge all its bridged modules will have to be reachable
        if self.device_type == DeviceType.NLE:
            # if there is a bridge it means it is a leaf
            if self.bridge:
                self.reachable = True
            elif self.modules:
                # this NLE is a bridge itself : make it not available
                self.reachable = False

        if not self.reachable and self.modules:
            # Update bridged modules and associated rooms
            for module_id in self.modules:
                module: Module = self.home.modules[module_id]
                await module.update(raw_data)
                if module.room_id:
                    self.home.rooms[module.room_id].update(raw_data)

    def update_features(self) -> None:
        """Update features."""

        self.features.update({var for var in vars(self) if var not in ATTRIBUTE_FILTER})
        if "battery_state" in vars(self) or "battery_percent" in vars(self):
            self.features.add("battery")
        if "wind_angle" in self.features:
            self.features.add("wind_direction")
            self.features.add("gust_direction")


class Camera(
    FirmwareMixin,
    MonitoringMixin,
    EventMixin,
    CameraMixinBase,
    WifiMixin,
    Module,
):
    """Class to represent a Netatmo camera."""

    async def update(self, raw_data: RawData) -> None:
        """Update camera with the latest data."""

        await Module.update(self, raw_data)
        await self.async_update_camera_urls()


class Switch(
    FirmwareMixin,
    EnergyHistoryMixin,
    PowerMixin,
    SwitchMixin,
    ApplianceTypeMixin,
    Module,
):
    """Class to represent a Netatmo switch."""


class Dimmer(DimmableMixin, Switch):
    """Class to represent a Netatmo dimmer."""


class Shutter(FirmwareMixin, ShutterMixin, Module):
    """Class to represent a Netatmo shutter."""


class Fan(FirmwareMixin, FanSpeedMixin, PowerMixin, Module):
    """Class to represent a Netatmo ventilation device."""


class RemoteControlMixin(FirmwareMixin, Module):
    """Class to represent a Netatmo remote control."""


class Energy(EnergyHistoryMixin, Module):
    """Class to represent a Netatmo energy module."""


class Boiler(BoilerMixin, Module):
    """Class to represent a Netatmo boiler."""



"""Common constants."""

from __future__ import annotations

from typing import Any

ERRORS: dict[int, str] = {
    400: "Bad request",
    401: "Unauthorized",
    403: "Forbidden",
    404: "Not found",
    406: "Not Acceptable",
    500: "Internal Server Error",
    502: "Bad Gateway",
    503: "Service Unavailable",
}

# Special types
RawData = dict[str, Any]

DEFAULT_BASE_URL: str = "https://api.netatmo.com/"
# Alternative base URL for commands restricted on the public OAuth2 API.
# app.netatmo.net accepts the same OAuth2 tokens but exposes endpoints
# (e.g. siren_status) that api.netatmo.com currently rejects.
SIREN_BASE_URL: str = "https://app.netatmo.net/"

# Endpoints
AUTH_REQ_ENDPOINT = "oauth2/token"
AUTH_URL_ENDPOINT = "oauth2/authorize"
WEBHOOK_URL_ADD_ENDPOINT = "api/addwebhook"
WEBHOOK_URL_DROP_ENDPOINT = "api/dropwebhook"

GETHOMESDATA_ENDPOINT = "api/homesdata"
GETHOMESTATUS_ENDPOINT = "api/homestatus"
GETEVENTS_ENDPOINT = "api/getevents"
SETTHERMMODE_ENDPOINT = "api/setthermmode"
SETROOMTHERMPOINT_ENDPOINT = "api/setroomthermpoint"
GETROOMMEASURE_ENDPOINT = "api/getroommeasure"
SWITCHHOMESCHEDULE_ENDPOINT = "api/switchhomeschedule"
SYNCHOMESCHEDULE_ENDPOINT = "api/synchomeschedule"

GETHOMEDATA_ENDPOINT = "api/gethomedata"
GETCAMERAPICTURE_ENDPOINT = "api/getcamerapicture"
GETEVENTSUNTIL_ENDPOINT = "api/geteventsuntil"
SETPERSONSAWAY_ENDPOINT = "api/setpersonsaway"
SETPERSONSHOME_ENDPOINT = "api/setpersonshome"
SETSTATE_ENDPOINT = "api/setstate"

GETHOMECOACHDATA_ENDPOINT = "api/gethomecoachsdata"

GETMEASURE_ENDPOINT = "api/getmeasure"
GETSTATIONDATA_ENDPOINT = "api/getstationsdata"

GETPUBLIC_DATA_ENDPOINT = "api/getpublicdata"

WEBRTC_OFFER_ENDPOINT = "api/webrtc/offer"
WEBRTC_TERMINATE_ENDPOINT = "api/webrtc/terminate"

AUTHORIZATION_HEADER = "Authorization"

# Possible scops
ALL_SCOPES: list[str] = [
    "access_camera",  # Netatmo camera products
    "access_doorbell",  # Netatmo Smart Video Doorbell
    "access_presence",  # Netatmo Smart Outdoor Camera
    "access_camerapro",  # Netatmo Indoor Camera Advance
    "read_bubendorff",  # Bubbendorf shutters
    "read_bfi",  # BTicino IP
    "read_camera",  # Netatmo camera products
    "read_carbonmonoxidedetector",  # Netatmo CO sensor
    "read_doorbell",  # Netatmo Smart Video Doorbell
    "read_homecoach",  # Netatmo Smart Indoor Air Quality Monitor
    "read_magellan",  # Legrand Wiring device or Electrical panel products
    "read_mhs1",  # Bticino MyHome Server 1 modules
    "read_mx",  # BTicino Classe 300 EOS
    "read_presence",  # Netatmo Smart Outdoor Camera
    "read_smarther",  # Smarther with Netatmo thermostat
    "read_smokedetector",  # Smart Smoke Alarm information and events
    "read_station",  # Netatmo weather station
    "read_thermostat",  # Netatmo climate products
    "read_camerapro",  # Netatmo Indoor Camera Advance
    "write_bubendorff",  # Bubbendorf shutters
    "write_bfi",  # BTicino IP
    "write_camera",  # Netatmo camera products
    "write_magellan",  # Legrand Wiring device or Electrical panel products
    "write_mhs1",  # Bticino MyHome Server 1 modules
    "write_mx",  # BTicino Classe 300 EOS
    "write_presence",  # Netatmo Smart Outdoor Camera
    "write_smarther",  # Smarther products
    "write_thermostat",  # Netatmo climate products
    "write_camerapro",  # Netatmo Indoor Camera Advance
]

EVENTS = "events"
SCHEDULES = "schedules"

MANUAL = "manual"
MAX = "max"
HOME = "home"
FROSTGUARD = "hg"
SCHEDULE = "schedule"
OFF = "off"
AWAY = "away"

HEATING = "heating"
COOLING = "cooling"
IDLE = "idle"


PILOT_WIRE_COMFORT = "comfort"
PILOT_WIRE_AWAY = "away"
PILOT_WIRE_FROST_GUARD = "frost_guard"
PILOT_WIRE_STAND_BY = "stand_by"
PILOT_WIRE_COMFORT_1 = "comfort_1"  # => doc unclear and contradictory: it is in the json schema but no in the doc
PILOT_WIRE_COMFORT_2 = "comfort_2"  # => same


STATION_TEMPERATURE_TYPE = "temperature"
STATION_PRESSURE_TYPE = "pressure"
STATION_HUMIDITY_TYPE = "humidity"

ACCESSORY_RAIN_LIVE_TYPE = "rain_live"
ACCESSORY_RAIN_60MIN_TYPE = "rain_60min"
ACCESSORY_RAIN_24H_TYPE = "rain_24h"
ACCESSORY_RAIN_TIME_TYPE = "rain_timeutc"
ACCESSORY_WIND_STRENGTH_TYPE = "wind_strength"
ACCESSORY_WIND_ANGLE_TYPE = "wind_angle"
ACCESSORY_WIND_TIME_TYPE = "wind_timeutc"
ACCESSORY_GUST_STRENGTH_TYPE = "gust_strength"
ACCESSORY_GUST_ANGLE_TYPE = "gust_angle"

# 2 days of dynamic historical data stored
MAX_HISTORY_TIME_FRAME: int = 24 * 2 * 3600

UNKNOWN = "unknown"

# Error codes
CONCURRENCY_ERROR_CODE = 11
THROTTLING_ERROR_CODE = 26
FORBIDDEN_ERROR_CODE = 403
TOO_MANY_REQUESTS_ERROR_CODE = 429

# Location constants
GPS_COORDINATES_COUNT = 2



"""Tests for pyatmo.auth retry / 429 concurrency handling."""

from __future__ import annotations

from datetime import UTC, datetime, timedelta
from email.utils import format_datetime

import pytest
from tenacity import Future, RetryCallState

from pyatmo.auth import (
    INITIAL_BACKOFF,
    MAX_BACKOFF,
    MAX_RETRIES,
    MAX_RETRY_AFTER,
    AbstractAsyncAuth,
    _parse_retry_after,
    _wait_retry_after,
)
from pyatmo.exceptions import ApiError, ApiThrottlingError, ApiTooManyRequestError

from .common import MockResponse


class _Auth(AbstractAsyncAuth):
    """Concrete auth for exercising error handling."""

    async def async_get_access_token(self) -> str:
        return "token"


@pytest.fixture
def auth():
    return _Auth(websession=None)


def _retry_state(exc: BaseException | None, attempt_number: int = 1) -> RetryCallState:
    """Build a minimal RetryCallState whose outcome raised ``exc``."""
    state = RetryCallState(retry_object=None, fn=None, args=(), kwargs={})
    state.attempt_number = attempt_number
    future = Future(attempt_number)
    if exc is not None:
        future.set_exception(exc)
    else:
        future.set_result(None)
    state.outcome = future
    return state


def test_too_many_request_error_carries_retry_after():
    """ApiTooManyRequestError exposes retry_after and is an ApiError."""
    err = ApiTooManyRequestError("boom", retry_after=12.0)

    assert isinstance(err, ApiError)
    assert err.retry_after == 12.0


def test_too_many_request_error_defaults_retry_after_none():
    """retry_after defaults to None when not provided."""
    assert ApiTooManyRequestError("boom").retry_after is None


def test_parse_retry_after_delta_seconds():
    """A numeric Retry-After header parses to float seconds."""
    assert _parse_retry_after("5") == 5.0
    assert _parse_retry_after(" 5 ") == 5.0
    assert _parse_retry_after("005") == 5.0


def test_parse_retry_after_empty_or_none():
    """Missing header yields None."""
    assert _parse_retry_after(None) is None
    assert _parse_retry_after("") is None


def test_parse_retry_after_http_date_future():
    """An HTTP-date in the future yields a positive delay."""
    future = datetime.now(UTC) + timedelta(seconds=30)
    value = format_datetime(future, usegmt=True)

    result = _parse_retry_after(value)

    assert result is not None
    assert 20 <= result <= 30


def test_parse_retry_after_http_date_past_clamped_to_zero():
    """An HTTP-date in the past clamps to 0."""
    past = datetime.now(UTC) - timedelta(seconds=30)
    value = format_datetime(past, usegmt=True)

    assert _parse_retry_after(value) == 0.0


def test_parse_retry_after_garbage():
    """Unparseable header yields None."""
    assert _parse_retry_after("not-a-date") is None


def test_wait_retry_after_uses_server_hint():
    """When the exception carries retry_after, that value drives the wait."""
    state = _retry_state(ApiTooManyRequestError("boom", retry_after=3.0))

    assert _wait_retry_after(state) == 3.0


def test_wait_retry_after_caps_server_hint():
    """A huge server hint is capped to MAX_RETRY_AFTER."""
    state = _retry_state(ApiTooManyRequestError("boom", retry_after=9999.0))

    assert _wait_retry_after(state) == MAX_RETRY_AFTER


def test_wait_retry_after_falls_back_without_hint():
    """No server hint -> bounded exponential fallback."""
    state = _retry_state(
        ApiTooManyRequestError("boom", retry_after=None),
        attempt_number=1,
    )

    result = _wait_retry_after(state)

    assert INITIAL_BACKOFF <= result <= MAX_BACKOFF + 1


def test_wait_retry_after_falls_back_for_other_exception():
    """A non-ApiTooManyRequestError outcome uses the fallback wait."""
    state = _retry_state(ApiError("other"))

    result = _wait_retry_after(state)

    assert INITIAL_BACKOFF <= result <= MAX_BACKOFF + 1


async def test_handle_error_429_code_11_raises_too_many_with_retry_after(auth):
    """429 + concurrency code 11 raises ApiTooManyRequestError w/ parsed hint."""
    resp = MockResponse(
        {"error": {"code": 11, "message": "concurrency"}},
        429,
        headers={"Retry-After": "7"},
    )

    with pytest.raises(ApiTooManyRequestError) as exc_info:
        await auth.handle_error_response(resp, 429, "https://x/y")

    assert exc_info.value.retry_after == 7.0


async def test_handle_error_429_code_11_without_header(auth):
    """No Retry-After header -> retry_after is None."""
    resp = MockResponse({"error": {"code": 11, "message": "concurrency"}}, 429)

    with pytest.raises(ApiTooManyRequestError) as exc_info:
        await auth.handle_error_response(resp, 429, "https://x/y")

    assert exc_info.value.retry_after is None


async def test_handle_error_403_code_26_raises_throttling(auth):
    """403 + code 26 raises ApiThrottlingError, not the 429 type."""
    resp = MockResponse({"error": {"code": 26, "message": "throttled"}}, 403)

    with pytest.raises(ApiThrottlingError):
        await auth.handle_error_response(resp, 403, "https://x/y")


async def test_handle_error_missing_error_key_no_keyerror(auth):
    """A body without an 'error' object raises ApiError, not KeyError."""
    resp = MockResponse({"status": "error"}, 500)

    with pytest.raises(ApiError):
        await auth.handle_error_response(resp, 500, "https://x/y")


async def test_post_api_request_retries_then_succeeds(auth):
    """A transient 429/code-11 is retried and the eventual success returned."""
    success = MockResponse({"status": "ok"}, 200)
    calls = 0

    busy = "busy"

    async def flaky(*_args, **_kwargs):
        nonlocal calls
        calls += 1
        if calls < 3:
            raise ApiTooManyRequestError(busy, retry_after=0.0)
        return success

    auth.async_post_request = flaky

    result = await auth.async_post_api_request(endpoint="api/homestatus")

    assert result is success
    assert calls == 3


async def test_post_api_request_reraises_after_exhaustion(auth):
    """Persistent 429/code-11 reraises ApiTooManyRequestError after MAX_RETRIES."""
    calls = 0
    busy = "busy"

    async def always_busy(*_args, **_kwargs):
        nonlocal calls
        calls += 1
        raise ApiTooManyRequestError(busy, retry_after=0.0)

    auth.async_post_request = always_busy

    with pytest.raises(ApiTooManyRequestError):
        await auth.async_post_api_request(endpoint="api/homestatus")

    assert calls == MAX_RETRIES


async def test_post_api_request_does_not_retry_other_errors(auth):
    """A generic ApiError is not retried."""
    calls = 0
    nope = "nope"

    async def boom(*_args, **_kwargs):
        nonlocal calls
        calls += 1
        raise ApiError(nope)

    auth.async_post_request = boom

    with pytest.raises(ApiError):
        await auth.async_post_api_request(endpoint="api/homestatus")

    assert calls == 1



"""Exceptions for pyatmo."""


class NoScheduleError(Exception):
    """Raised when no schedule is found."""


class InvalidScheduleError(Exception):
    """Raised when an invalid schedule is encountered."""


class InvalidHomeError(Exception):
    """Raised when an invalid home is encountered."""


class InvalidRoomError(Exception):
    """Raised when an invalid room is encountered."""


class NoDeviceError(Exception):
    """Raised when no device is found."""


class ApiError(Exception):
    """Raised when an API error is encountered."""


class ApiThrottlingError(ApiError):
    """Raised when an API error is encountered."""


class ApiHomeReachabilityError(ApiError):
    """Raised when an API error is encountered."""


class InvalidStateError(Exception):
    """Raised when an invalid state is encountered."""


class ApiTooManyRequestError(ApiError):
    """Raised when API returned 429 code 11."""

    def __init__(self, message: str, retry_after: float | None = None) -> None:
        """Initialize with an optional server-provided retry delay (seconds)."""
        super().__init__(message)
        self.retry_after = retry_after



"""Base class for Netatmo entities."""

from __future__ import annotations

from abc import ABC
import bisect
from dataclasses import dataclass
import logging
from operator import itemgetter
from time import time
from typing import TYPE_CHECKING, Any

from pyatmo.const import GPS_COORDINATES_COUNT, MAX_HISTORY_TIME_FRAME, RawData
from pyatmo.event import EventTypes
from pyatmo.modules.device_types import (
    ApplianceType,
    BoilerControl,
    BoilerError,
    DeviceType,
    DhwControl,
    DoorTagCategory,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from pyatmo.home import Home
    from pyatmo.modules.module import ModuleT


LOG: logging.Logger = logging.getLogger(__name__)


def bridged_module_ids(raw_data: dict[str, Any], default: Any = None) -> Any:  # noqa: ANN401
    """Return the bridged-children module ids.

    The /homesdata schema documents this list as `module_bridged` in some
    variants and `modules_bridged` in others; the API spelling is unconfirmed.
    Both are read, and `modules_bridged` takes precedence when both are present.
    """
    return raw_data.get("modules_bridged", raw_data.get("module_bridged", default))


NETATMO_ATTRIBUTES_MAP: dict[str, Callable[[dict[str, Any], Any], Any]] = {
    "entity_id": lambda x, y: x.get("id", y),
    "modules": bridged_module_ids,
    "device_type": lambda x, y: DeviceType(x.get("type", y)),
    "event_type": lambda x, y: EventTypes(x.get("type", y)),
    "reachable": lambda x, _: x.get("reachable", False),
    "monitoring": lambda x, _: x.get("monitoring", False) == "on",
    "battery_level": lambda x, _: x.get("battery_vp", x.get("battery_level")),
    "place": lambda x, y: Place(x["place"]) if isinstance(x.get("place"), dict) else y,
    "target_position__step": lambda x, _: x.get("target_position:step"),
    "appliance_type": lambda x, y: ApplianceType(x.get("appliance_type", y)),
    "doortag_category": lambda x, y: DoorTagCategory(x.get("category", y)),
    # Coerce only when present so absent fields keep their default (no spurious
    # "unknown" warning from Enum(None) on every non-OTH update).
    "boiler_control": lambda x, y: (
        BoilerControl(x["boiler_control"]) if "boiler_control" in x else y
    ),
    "boiler_error": lambda x, y: (
        BoilerError(x["boiler_error"]) if "boiler_error" in x else y
    ),
    "dhw_control": lambda x, y: (
        DhwControl(x["dhw_control"]) if "dhw_control" in x else y
    ),
}


def default(key: str, val: Any) -> Callable[[dict[str, Any], Any], Any]:  # noqa: ANN401
    """Return default value."""

    return lambda x, _: x.get(key, val)


def update_name(name: str, pre_fix: str) -> str:
    """Remove duplicates from string."""

    if name.startswith(pre_fix):
        return name
    return f"{pre_fix} {name}"


class EntityBase:
    """Base class for Netatmo entities."""

    entity_id: str
    home: Home
    bridge: str | None
    history_features: set[str]
    history_features_values: dict
    name: str

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize the entity."""
        self.home = home
        self.entity_id = module.get("id", "")
        self.name = module.get("name", f"Unknown {self.entity_id}")
        self.bridge = None
        self.history_features = set()
        self.history_features_values = {}

    def has_feature(self, feature: str) -> bool:
        """Check if the entity has the given feature."""

        return hasattr(self, feature) or feature in self.history_features


class NetatmoBase(EntityBase, ABC):
    """Base class for Netatmo entities."""

    def __init__(self, raw_data: RawData) -> None:
        """Initialize a Netatmo entity."""

        self.entity_id = raw_data["id"]
        self.name = raw_data.get("name", f"Unknown {self.entity_id}")
        self.history_features_values = {}
        self.history_features = set()

    def update_topology(self, raw_data: RawData) -> None:
        """Update topology."""

        self._update_attributes(raw_data)

    def _update_attributes(self, raw_data: RawData) -> None:
        """Update attributes."""

        self.__dict__ = {
            key: NETATMO_ATTRIBUTES_MAP.get(key, default(key, val))(raw_data, val)
            for key, val in self.__dict__.items()
        }

        now = int(time())
        for hist_feature in self.history_features:
            if hist_feature in self.__dict__:
                val = getattr(self, hist_feature)
                if val is None:
                    continue

                self.add_history_data(hist_feature, val, now)

    def add_history_data(
        self,
        feature: str,
        value: Any,  # noqa: ANN401
        time: int,
    ) -> None:
        """Add historical data at the given time."""

        # get the feature values rolling buffer
        hist_f = self.history_features_values.setdefault(feature, [])
        if not hist_f or hist_f[-1][0] <= time:
            hist_f.append((time, value, self.entity_id))
        else:
            i = bisect.bisect_left(hist_f, time, key=itemgetter(0))

            if i < len(hist_f) and hist_f[i][0] == time:
                hist_f[i] = (time, value, self.entity_id)
            else:
                hist_f.insert(i, (time, value, self.entity_id))

        # keep timing history to a maximum representative time
        while len(hist_f) > 0 and hist_f[-1][0] - hist_f[0][0] > MAX_HISTORY_TIME_FRAME:
            hist_f.pop(0)

    def get_history_data(
        self,
        feature: str,
        from_ts: float,
        to_ts: float | None = None,
    ) -> list[Any]:
        """Retrieve historical data."""

        hist_f = self.history_features_values.get(feature, [])

        if not hist_f:
            return []

        in_s: int = bisect.bisect_left(hist_f, from_ts, key=itemgetter(0))

        if to_ts is None:
            out_s: int = len(hist_f)
        else:
            out_s = bisect.bisect_right(hist_f, to_ts, key=itemgetter(0))

        return hist_f[in_s:out_s]


@dataclass
class Location:
    """Class of Netatmo public weather location."""

    latitude: float
    longitude: float

    def __init__(self, longitude: float, latitude: float) -> None:
        """Initialize self."""

        self.latitude = latitude
        self.longitude = longitude

    def __iter__(self) -> Iterator[float]:
        """Iterate over latitude and longitude."""

        yield self.longitude
        yield self.latitude


@dataclass
class Place:
    """Class of Netatmo public weather place."""

    altitude: int | None
    city: str | None
    country: str | None
    timezone: str | None
    location: Location | None

    def __init__(
        self,
        data: dict[str, Any] | None,
    ) -> None:
        """Initialize self."""

        self.altitude = None
        self.city = None
        self.country = None
        self.timezone = None
        self.location = None

        if not isinstance(data, dict):
            LOG.debug("Place data is unknown")
            return

        self.altitude = data.get("altitude")
        self.city = data.get("city")
        self.country = data.get("country")
        self.timezone = data.get("timezone")

        try:
            location_data: list[float] = list(data.get("location") or [])
        except TypeError:
            location_data = []

        if len(location_data) != GPS_COORDINATES_COUNT:
            LOG.debug("Invalid location data: %s", data)
            return

        self.location = Location(location_data[0], location_data[1])
